
/*******************************************************************************
*
* File check5.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Statistical check of random_asrc().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "sl3.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "aflds.h"
#include "global.h"

#define NCFG 1000
#define N0 (NPROC0*L0)

extern void plaq1(sl3_alg_dble *ad,double **plaq);
extern void plaq2(double **plaq);
extern void star1(sl3_alg_dble *ad,double *star);
extern void star2(double *star);

static int opv[(int)(PM_TYPES)];
static double *pv[(int)(PM_TYPES)];
static double **plaq,smp0[6][N0],smp1[6][N0];
static double *star,sms0[N0],sms1[N0];


static void new_fld(sl3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),16*(4*VOLUME+7*(BNDRY/4)+3));
   else
      gauss_dble((double*)(ad),16*(4*VOLUME+7*(BNDRY/4)));
}


static void alloc_plaq(void)
{
   int n;

   plaq=malloc(6*sizeof(*plaq));
   error(plaq==NULL,1,"alloc_plaq [check5.c]",
         "Unable to allocate plaquette array");
   plaq[0]=malloc(6*N0*sizeof(*plaq[0]));
   error(plaq[0]==NULL,2,"alloc_plaq [check5.c]",
         "Unable to allocate plaquette array");

   for (n=1;n<6;n++)
      plaq[n]=plaq[n-1]+N0;

   star=malloc(N0*sizeof(*star));
   error(star==NULL,1,"alloc_plaq [check5.c]",
         "Unable to allocate action density array");
}


static void default_pv(int order)
{
   int i,pmax;
   double *p;

   pmax=(int)(PM_TYPES);
   p=malloc(pmax*(order+1)*sizeof(*p));

   for (i=0;i<(pmax*(order+1));i++)
      p[i]=0.0;

   for (i=0;i<pmax;i++)
   {
      opv[i]=order;
      pv[i]=p;
      p+=order+1;
   }

   pv[(int)(C0)][0]=1.0;
   pv[(int)(LAMBDA)][0]=1.0;
   pv[(int)(MASS)][0]=0.0;
   pv[(int)(CG)][0]=1.0;
   pv[(int)(CGP)][0]=1.0;
   pv[(int)(CSW)][0]=1.0;
   pv[(int)(CF)][0]=1.0;
   pv[(int)(CFP)][0]=1.0;
}


int main(int argc,char *argv[])
{
   int my_rank,bc,order;
   int m,n,i,x0;
   double lambda0;
   double d,dmx,r1,sm0,sm1,sig[4];
   complex_dble z0,z1;
   sl3_alg_dble **wad;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check5.log","w",stdout);

      printf("\n");
      printf("Statistical check of random_asrc()\n");
      printf("----------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      error_root(bc==0,1,"main [check5.c]","Syntax: check5 [-bc <type>]");
      error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check5.c]",
                 "Syntax: check5 [-bc <type>]");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   order=1;
   lambda0=0.4567;
   set_th_parms(LYM,bc,0);
   default_pv(order);
   pv[(int)(LAMBDA)][0]=lambda0;
   set_ispt_parms(order,0,opv,pv);
   print_th_parms();
   print_ispt_parms();
   start_ranlux(0,1234);
   geometry();
   alloc_wad(3);
   alloc_wcd(1);
   alloc_plaq();

   wad=reserve_wad(3);
   new_fld(wad[0]);
   set_adbc(wad[0]);
   aprop(wad[0],wad[1]);
   z0=scalar_prod_calg(4*VOLUME,1,wad[0],wad[1]);
   sm0=0.0;
   sm1=0.0;

   for (n=0;n<6;n++)
   {
      for (x0=0;x0<N0;x0++)
      {
         smp0[n][x0]=0.0;
         smp1[n][x0]=0.0;
      }
   }

   for (x0=0;x0<N0;x0++)
   {
      sms0[x0]=0.0;
      sms1[x0]=0.0;
   }

   for (i=0;i<NCFG;i++)
   {
      random_asrc(wad[2]);
      z1=scalar_prod_calg(4*VOLUME,1,wad[2],wad[0]);
      r1=z1.re*z1.re+z1.im*z1.im;
      sm0+=r1;
      sm1+=r1*r1;

      plaq1(wad[2],plaq);
      star1(wad[2],star);

      for (n=0;n<6;n++)
      {
         for (x0=0;x0<N0;x0++)
         {
            smp0[n][x0]+=plaq[n][x0];
            smp1[n][x0]+=plaq[n][x0]*plaq[n][x0];
         }
      }

      for (x0=0;x0<N0;x0++)
      {
         sms0[x0]+=star[x0];
         sms1[x0]+=star[x0]*star[x0];
      }
   }

   sm0/=(double)(NCFG);
   sm1/=(double)(NCFG);
   sm1-=sm0*sm0;

   for (n=0;n<6;n++)
   {
      for (x0=0;x0<N0;x0++)
      {
         smp0[n][x0]/=(double)(NCFG);
         smp1[n][x0]/=(double)(NCFG);
         smp1[n][x0]-=smp0[n][x0]*smp0[n][x0];
         smp1[n][x0]=sqrt(smp1[n][x0]/(double)(NCFG));
      }
   }

   for (x0=0;x0<N0;x0++)
   {
      sms0[x0]/=(double)(NCFG);
      sms1[x0]/=(double)(NCFG);
      sms1[x0]-=sms0[x0]*sms0[x0];
      sms1[x0]=sqrt(sms1[x0]/(double)(NCFG));
   }

   plaq2(plaq);
   star2(star);

   if (my_rank==0)
   {
      printf("Check of (J,<AA>J):\n");
      printf("Number of configurations = %d\n",NCFG);
      printf("Expected value = %.4e (imaginary part = %.1e)\n",
             z0.re,fabs(z0.im));
      printf("Stochastically estimated value = %.4e+-%.1e\n\n",
             sm0,sqrt(sm1/(double)(NCFG)));

      printf("Check of plaquette values:\n");

      for (i=0;i<4;i++)
         sig[i]=0.0;

      m=0;
      dmx=0.0;

      for (n=0;n<6;n++)
      {
         for (x0=0;x0<N0;x0++)
         {
            printf("n,x0 = %d,%3d, expected = %.5e, estimated = %.5e+-%.1e\n",
                   n,x0,plaq[n][x0],smp0[n][x0],smp1[n][x0]);

            if (plaq[n][x0]>0.0)
            {
               d=fabs(smp0[n][x0]-plaq[n][x0])/smp1[n][x0];
               i=(int)(d);

               if (d>dmx)
                  dmx=d;

               if (i<3)
                  sig[i]+=1.0;
               else
                  sig[3]+=1.0;

               m+=1;
            }
            else if (smp0[n][x0]!=0.0)
               error_root(1,1,"main [check5.c]","Unexpected plaquette value");
         }
      }

      for (i=0;i<4;i++)
         sig[i]*=(100.0/(double)(m));

      printf("\n");
      printf("%.1f%%<1s, %.1f%%<2s, %.1f%%<3s, %.1f%%>=3s, max=%.1fs\n\n",
             sig[0],sig[1],sig[2],sig[3],dmx);

      printf("Check of the gauge-fixing density:\n");

      for (i=0;i<4;i++)
         sig[i]=0.0;

      m=0;
      dmx=0;

      for (x0=0;x0<N0;x0++)
      {
         printf("x0 = %3d, expected = %.5e, estimated = %.5e+-%.1e\n",
                x0,star[x0],sms0[x0],sms1[x0]);

         if (star[x0]!=0.0)
         {
            d=fabs(sms0[x0]-star[x0])/sms1[x0];
            i=(int)(d);

            if (d>dmx)
               dmx=d;

            if (i<3)
               sig[i]+=1.0;
            else
               sig[3]+=1.0;

            m+=1;
         }
         else if (sms0[x0]!=0.0)
            error_root(1,1,"main [check5.c]","Unexpected action density value");
      }

      for (i=0;i<4;i++)
         sig[i]*=(100.0/(double)(m));

      printf("\n");
      printf("%.1f%%<1s, %.1f%%<2s, %.1f%%<3s, %.1f%%>=3s, max=%.1fs\n\n",
             sig[0],sig[1],sig[2],sig[3],dmx);

      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
