
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Exporting and importing gauge potentials.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "archive.h"
#include "global.h"

#define NMAX 4


static void random_afld(sl3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),16*(4*VOLUME+7*(BNDRY/4)+3));
   else
      gauss_dble((double*)(ad),16*(4*VOLUME+7*(BNDRY/4)));

   set_adbc(ad);
}


static int is_equal(sl3_alg_dble *ad1,sl3_alg_dble *ad2)
{
   int ie;

   ie=1;
   ie&=((*ad1).c1.re==(*ad2).c1.re);
   ie&=((*ad1).c1.im==(*ad2).c1.im);
   ie&=((*ad1).c2.re==(*ad2).c2.re);
   ie&=((*ad1).c2.im==(*ad2).c2.im);
   ie&=((*ad1).c3.re==(*ad2).c3.re);
   ie&=((*ad1).c3.im==(*ad2).c3.im);
   ie&=((*ad1).c4.re==(*ad2).c4.re);
   ie&=((*ad1).c4.im==(*ad2).c4.im);
   ie&=((*ad1).c5.re==(*ad2).c5.re);
   ie&=((*ad1).c5.im==(*ad2).c5.im);
   ie&=((*ad1).c6.re==(*ad2).c6.re);
   ie&=((*ad1).c6.im==(*ad2).c6.im);
   ie&=((*ad1).c7.re==(*ad2).c7.re);
   ie&=((*ad1).c7.im==(*ad2).c7.im);
   ie&=((*ad1).c8.re==(*ad2).c8.re);
   ie&=((*ad1).c8.im==(*ad2).c8.im);

   return ie;
}


static int cmp_ad(sl3_alg_dble *ad1,sl3_alg_dble *ad2)
{
   int ie;
   sl3_alg_dble *adm;

   adm=ad1+4*VOLUME;
   ie=1;

   for (;ad1<adm;ad1++)
   {
      ie&=is_equal(ad1,ad2);
      ad2+=1;
   }

   return ie;
}


int main(int argc,char *argv[])
{
   int my_rank,bc,ie;
   int n,i,nsize;
   sl3_alg_dble **wad;
   char afld_dir[NAME_SIZE],name[NAME_SIZE];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   set_error_file("check2.log",0);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);
      fin=freopen("check2.in","r",stdin);

      printf("\n");
      printf("Exporting and importing momentum fields\n");
      printf("---------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("afld_dir","%s\n",afld_dir);
      fclose(fin);

      bc=find_opt(argc,argv,"-bc");
      error_root((bc==0)||(sscanf(argv[bc+1],"%d",&bc)!=1),1,
                 "main [check2.c]","Syntax: check2 -bc <type>");
   }

   MPI_Bcast(afld_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();

   check_dir_root(afld_dir);
   nsize=name_size("%s/testafldn%d",afld_dir,NMAX);
   error_root(nsize>=NAME_SIZE,1,"main [check2.c]","afld_dir name is too long");

   alloc_wad(2*NMAX);
   alloc_wud(NMAX);

   if (my_rank==0)
   {
      printf("Export random gauge potentials to the files %s/testafldn*\n"
             "on the local disks. Then read the fields from there\n",afld_dir);
      printf("and compare with the saved fields.\n\n");
   }

   for (n=0;n<NMAX;n++)
   {
      wad=reserve_wad(2*(n+1));

      for (i=0;i<=n;i++)
      {
         random_afld(wad[i]);
         assign_calg2calg(4*VOLUME+7*(BNDRY/4),wad[i],wad[i+n+1]);
      }

      sprintf(name,"%s/testafldn%d",afld_dir,n);
      export_afld(name,n,wad);

      for (i=0;i<=n;i++)
         random_afld(wad[i]);

      import_afld(name,n,wad);
      ie=1;

      for (i=0;i<=n;i++)
      {
         ie&=check_adbc(wad[i]);
         ie&=cmp_ad(wad[i],wad[i+n+1]);
      }

      error(ie!=1,1,"main [check2.c]",
            "The gauge potential is not properly restored");
      release_wad();
   }

   if (my_rank==0)
   {
      printf("No errors detected --- the fields are correctly exported\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
