
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Action of the Faddeev-Popov operator on plane waves.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "pmdb.h"
#include "utils.h"
#include "random.h"
#include "lattice.h"
#include "linalg.h"
#include "cflds.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)


static sl3_alg_dble mul_cplx(complex_dble z,sl3_alg_dble s)
{
   sl3_alg_dble r;

   r.c1.re=z.re*s.c1.re-z.im*s.c1.im;
   r.c1.im=z.im*s.c1.re+z.re*s.c1.im;
   r.c2.re=z.re*s.c2.re-z.im*s.c2.im;
   r.c2.im=z.im*s.c2.re+z.re*s.c2.im;
   r.c3.re=z.re*s.c3.re-z.im*s.c3.im;
   r.c3.im=z.im*s.c3.re+z.re*s.c3.im;
   r.c4.re=z.re*s.c4.re-z.im*s.c4.im;
   r.c4.im=z.im*s.c4.re+z.re*s.c4.im;
   r.c5.re=z.re*s.c5.re-z.im*s.c5.im;
   r.c5.im=z.im*s.c5.re+z.re*s.c5.im;
   r.c6.re=z.re*s.c6.re-z.im*s.c6.im;
   r.c6.im=z.im*s.c6.re+z.re*s.c6.im;
   r.c7.re=z.re*s.c7.re-z.im*s.c7.im;
   r.c7.im=z.im*s.c7.re+z.re*s.c7.im;
   r.c8.re=z.re*s.c8.re-z.im*s.c8.im;
   r.c8.im=z.im*s.c8.re+z.re*s.c8.im;

   return r;
}


int main(int argc,char *argv[])
{
   int my_rank,bc;
   int n,i,x0,x1,x2,x3,ix;
   int np[4],bo[4];
   double pi,d,dmax,phsq,px0,px1,rt;
   double ran[4],p[4];
   complex_dble z;
   sl3_alg_dble **wcd,c0,c1,A,B;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);
      printf("\n");
      printf("Action of the Faddeev-Popov operator on plane waves\n");
      printf("---------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check2.c]",
                    "Syntax: check2 [-bc <type>]");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();
   alloc_wcd(3);
   wcd=reserve_wcd(3);

   pi=4.0*atan(1.0);
   n=10;
   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;
   dmax=0.0;

   for (i=0;i<n;i++)
   {
      ranlxd(ran,4);

      np[0]=(int)(ran[0]*(double)(N0));
      np[1]=(int)(ran[1]*(double)(N1));
      np[2]=(int)(ran[2]*(double)(N2));
      np[3]=(int)(ran[3]*(double)(N3));

      if (bc==1)
      {
         if (np[0]==0)
            np[0]=1;
         p[0]=(double)(np[0])*pi/(double)(N0);
      }
      else
         p[0]=((double)(np[0])+0.5)*pi/(double)(N0);

      p[1]=(double)(np[1])*2.0*pi/(double)(N1);
      p[2]=(double)(np[2])*2.0*pi/(double)(N2);
      p[3]=(double)(np[3])*2.0*pi/(double)(N3);

      gauss_dble((double*)(&A),16);
      gauss_dble((double*)(&B),16);

      MPI_Bcast(p,4,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&A,16,MPI_DOUBLE,0,MPI_COMM_WORLD);
      MPI_Bcast(&B,16,MPI_DOUBLE,0,MPI_COMM_WORLD);

      phsq=(1.0-cos(p[0]));
      phsq+=(1.0-cos(p[1]));
      phsq+=(1.0-cos(p[2]));
      phsq+=(1.0-cos(p[3]));
      phsq*=2.0;

      for (x0=0;x0<L0;x0++)
      {
         for (x1=0;x1<L1;x1++)
         {
            for (x2=0;x2<L2;x2++)
            {
               for (x3=0;x3<L3;x3++)
               {
                  ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];

                  rt=1.0-(double)(x0+bo[0])/(double)(N0);
                  _sl3_alg_mulr_assign(c0,rt,A);

                  px0=p[0]*(double)(x0+bo[0]);
                  px1=p[1]*(double)(x1+bo[1]);
                  px1+=p[2]*(double)(x2+bo[2]);
                  px1+=p[3]*(double)(x3+bo[3]);

                  if (bc==1)
                  {
                     z.re=sin(px0)*cos(px1);
                     z.im=sin(px0)*sin(px1);
                  }
                  else
                  {
                     z.re=cos(px0)*cos(px1);
                     z.im=cos(px0)*sin(px1);
                  }

                  c1=mul_cplx(z,B);

                  if (bc==1)
                  {
                     wcd[0][ix]=c0;
                     _sl3_alg_add_assign(wcd[0][ix],c1);
                  }
                  else
                     wcd[0][ix]=c1;

                  rt=rt/(double)(N0*N0);
                  _sl3_alg_mulr_assign(c0,rt,A);

                  z.re*=phsq;
                  z.im*=phsq;
                  c1=mul_cplx(z,B);

                  if (bc==1)
                  {
                     wcd[1][ix]=c0;
                     _sl3_alg_add_assign(wcd[1][ix],c1);
                  }
                  else
                     wcd[1][ix]=c1;
               }
            }
         }
      }

      dfp(wcd[0],wcd[2]);
      mulr_add_calg(VOLUME,wcd[2],wcd[1],-1.0);
      d=norm_square_calg(VOLUME,1,wcd[2])/norm_square_calg(VOLUME,1,wcd[0]);
      d=sqrt(d);
      if (d>dmax)
         dmax=d;

      if (my_rank==0)
         printf("Normalized deviation = %.1e at p=(%d,%d,%d,%d)\n",
                d,np[0],np[1],np[2],np[3]);
   }

   if (my_rank==0)
   {
      printf("\n");
      printf("Maximal normalized deviation = %.1e\n\n",dmax);
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
