
/*******************************************************************************
*
* File check4.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Statistical check of random_csrc().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "cflds.h"
#include "global.h"

#define NCFG 1000
#define N0 (NPROC0*L0)

extern void cnrm1(sl3_alg_dble *cd1,sl3_alg_dble *cd2,complex_dble *cnrm);
extern void cnrm2(double *cnrm);
extern void star1(sl3_alg_dble *cd1,sl3_alg_dble *cd2,complex_dble **star);
extern void star2(double **star);

static double *cnrm0,**star0;
static double smc1[N0],sms1[4][N0];
static complex_dble *cnrm,**star;
static complex_dble smc0[N0],sms0[4][N0];


static void new_fld(sl3_alg_dble *cd)
{
   gauss_dble((double*)(cd),16*NSPIN);
}


static void alloc_obs(void)
{
   int mu;

   star=malloc(4*sizeof(*star));
   error(star==NULL,1,"alloc_obs [check4.c]",
         "Unable to allocate star array");
   star[0]=malloc(4*N0*sizeof(*star[0]));
   error(star[0]==NULL,2,"alloc_obs [check4.c]",
         "Unable to allocate star array");

   for (mu=1;mu<4;mu++)
      star[mu]=star[mu-1]+N0;

   cnrm=malloc(N0*sizeof(*cnrm));
   error(cnrm==NULL,1,"alloc_obs [check4.c]",
         "Unable to allocate cnrm array");

   star0=malloc(4*sizeof(*star0));
   error(star0==NULL,1,"alloc_obs [check4.c]",
         "Unable to allocate star0 array");
   star0[0]=malloc(4*N0*sizeof(*star0[0]));
   error(star0[0]==NULL,2,"alloc_obs [check4.c]",
         "Unable to allocate star0 array");

   for (mu=1;mu<4;mu++)
      star0[mu]=star0[mu-1]+N0;

   cnrm0=malloc(N0*sizeof(*cnrm0));
   error(cnrm0==NULL,1,"alloc_obs [check4.c]",
         "Unable to allocate cnrm0 array");
}


int main(int argc,char *argv[])
{
   int my_rank,bc;
   int n,i,mu,x0;
   double sm0,sm1,sig[4];
   complex_dble z0,z1,z2,z3;
   sl3_alg_dble **wcd;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check4.log","w",stdout);

      printf("\n");
      printf("Statistical check of random_csrc()\n");
      printf("----------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check4.c]",
                    "Syntax: check4 [-bc <type>]");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,654);
   geometry();
   alloc_wad(2);
   alloc_wcd(4);
   alloc_obs();

   wcd=reserve_wcd(4);
   new_fld(wcd[0]);
   set_cdbc(wcd[0]);
   cprop(wcd[0],wcd[1]);
   z0=scalar_prod_calg(VOLUME,1,wcd[0],wcd[1]);
   sm0=0.0;
   sm1=0.0;

   for (mu=0;mu<4;mu++)
   {
      for (x0=0;x0<N0;x0++)
      {
         sms0[mu][x0].re=0.0;
         sms0[mu][x0].im=0.0;
         sms1[mu][x0]=0.0;
      }
   }

   for (x0=0;x0<N0;x0++)
   {
      smc0[x0].re=0.0;
      smc0[x0].im=0.0;
      smc1[x0]=0.0;
   }

   for (i=0;i<NCFG;i++)
   {
      random_csrc(wcd[3],wcd[2]);
      z1=scalar_prod_calg(VOLUME,1,wcd[0],wcd[2]);
      assign_calgdag2calg(VOLUME,wcd[3],wcd[3]);
      z2=scalar_prod_calg(VOLUME,1,wcd[3],wcd[0]);
      assign_calgdag2calg(VOLUME,wcd[3],wcd[3]);
      z3.re=z1.re*z2.re-z1.im*z2.im;
      z3.im=z1.re*z2.im+z1.im*z2.re;
      sm0-=z3.re;
      sm1+=(z3.re*z3.re+z3.im*z3.im);

      star1(wcd[2],wcd[3],star);
      cnrm1(wcd[2],wcd[3],cnrm);

      for (mu=0;mu<4;mu++)
      {
         for (x0=0;x0<N0;x0++)
         {
            sms0[mu][x0].re+=star[mu][x0].re;
            sms0[mu][x0].im+=star[mu][x0].im;
            sms1[mu][x0]+=(star[mu][x0].re*star[mu][x0].re+
                           star[mu][x0].im*star[mu][x0].im);
         }
      }

      for (x0=0;x0<N0;x0++)
      {
         smc0[x0].re+=cnrm[x0].re;
         smc0[x0].im+=cnrm[x0].im;
         smc1[x0]+=(cnrm[x0].re*cnrm[x0].re+cnrm[x0].im*cnrm[x0].im);
      }
   }

   sm0/=(double)(NCFG);
   sm1/=(double)(NCFG);
   sm1-=sm0*sm0;
   sm1=sqrt(sm1/(double)(NCFG));

   for (mu=0;mu<4;mu++)
   {
      for (x0=0;x0<N0;x0++)
      {
         sms0[mu][x0].re/=(double)(NCFG);
         sms0[mu][x0].im/=(double)(NCFG);
         sms1[mu][x0]/=(double)(NCFG);
         sms1[mu][x0]-=(sms0[mu][x0].re*sms0[mu][x0].re+
                        sms0[mu][x0].im*sms0[mu][x0].im);
         sms1[mu][x0]=sqrt(sms1[mu][x0]/(double)(NCFG));
      }
   }

   for (x0=0;x0<N0;x0++)
   {
      smc0[x0].re/=(double)(NCFG);
      smc0[x0].im/=(double)(NCFG);
      smc1[x0]/=(double)(NCFG);
      smc1[x0]-=(smc0[x0].re*smc0[x0].re+smc0[x0].im*smc0[x0].im);
      smc1[x0]=sqrt(smc1[x0]/(double)(NCFG));
   }

   star2(star0);
   cnrm2(cnrm0);

   if (my_rank==0)
   {
      printf("Check of (J,<c*cbar>J):\n");
      printf("Number of configurations = %d\n",NCFG);
      printf("Expected value = %.4e (imaginary part = %.1e)\n",
             z0.re,fabs(z0.im));
      printf("Stochastically estimated value = %.4e+-%.1e\n\n",
             sm0,sm1);

      printf("Check of star values:\n");

      for (i=0;i<4;i++)
         sig[i]=0.0;

      n=0;

      for (mu=0;mu<4;mu++)
      {
         for (x0=0;x0<N0;x0++)
         {
            printf("mu,x0 = %d,%3d, expected = %.5e, estimated = %.5e+-%.1e, "
                   "imaginary part = %.1e\n",mu,x0,star0[mu][x0],
                   sms0[mu][x0].re,sms1[mu][x0],fabs(sms0[mu][x0].im));

            if (star0[mu][x0]!=0.0)
            {
               i=(int)(fabs(sms0[mu][x0].re-star0[mu][x0])/sms1[mu][x0]);

               if (i<3)
                  sig[i]+=1.0;
               else
                  sig[3]+=1.0;

               n+=1;
            }
            else if (sms0[mu][x0].re!=0.0)
               error(1,1,"main [check4.c]","Unexpected star value");
         }
      }

      for (i=0;i<4;i++)
         sig[i]*=(100.0/(double)(n));

      printf("\n");
      printf("%.1e%%<1s, %.1e%%<2s, %.1e%%<3s, %.1e%%>=3s\n\n",
             sig[0],sig[1],sig[2],sig[3]);

      printf("Check of time-slice norm:\n");

      for (i=0;i<4;i++)
         sig[i]=0.0;

      n=0;

      for (x0=0;x0<N0;x0++)
      {
         printf("x0 = %3d, expected = %.5e, estimated = %.5e+-%.1e, "
                "imaginary part = %.1e\n",
                x0,cnrm0[x0],smc0[x0].re,smc1[x0],fabs(smc0[x0].im));

         if (cnrm0[x0]!=0.0)
         {
            i=(int)(fabs(smc0[x0].re-cnrm0[x0])/smc1[x0]);

            if (i<3)
               sig[i]+=1.0;
            else
               sig[3]+=1.0;

            n+=1;
         }
         else if (smc0[x0].re!=0.0)
            error(1,1,"main [check4.c]","Unexpected time-slice norm value");
      }

      for (i=0;i<4;i++)
         sig[i]*=(100.0/(double)(n));

      printf("\n");
      printf("%.1e%%<1s, %.1e%%<2s, %.1e%%<3s, %.1e%%>=3s\n",
             sig[0],sig[1],sig[2],sig[3]);
   }

   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
