
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Direct check of ispt().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "mpi.h"
#include "pmdb.h"
#include "utils.h"
#include "random.h"
#include "lattice.h"
#include "tupel.h"
#include "linalg.h"
#include "aflds.h"
#include "cflds.h"
#include "vxfcts.h"
#include "ispt.h"
#include "global.h"

static int my_rank,bc,Nf,order,iroot,ncfg;
static int opv[(int)(PM_TYPES)];
static double lambda,*pv[(int)(PM_TYPES)];
static theory_t th;
static char theory[NAME_SIZE];
static FILE *fin=NULL,*flog=NULL;


static void set_th(void)
{
   if (strcmp("PHI4",theory)==0)
      th=PHI4;
   else if (strcmp("YM",theory)==0)
      th=YM;
   else if (strcmp("QCD",theory)==0)
      th=QCD;
   else if (strcmp("LYM",theory)==0)
      th=LYM;
   else if (strcmp("LQCD",theory)==0)
      th=LQCD;
   else
      error(1,1,"set_th [check3.c]","Unknown theory");
}


static void read_theory_parms(void)
{
   find_section("Theory");
   read_line("theory","%s",theory);
   read_line("bc","%d",&bc);
   read_line("Nf","%d",&Nf);
   set_th();

   find_section("ISPT");
   read_line("order","%d",&order);
   read_line("iroot","%d",&iroot);
   read_line("lambda","%lf",&lambda);

   find_section("Run");
   read_line("ncfg","%d",&ncfg);

   error_root((th!=YM)&&(th!=LYM),1,"read_theory_parms [check3.c]",
              "Currently the theory must be YM or LYM");

   error_root((order<1)||(order>2),1,"read_theory_parms [check3.c]",
              "The order in the coupling must be 1 or 2");
}


static void read_infile(void)
{
   int ith;

   if (my_rank==0)
   {
      fin=freopen("check3.in","r",stdin);
      error_root(fin==NULL,1,"read_infile [check3.c]",
                 "Unable to open input file");
      read_theory_parms();
      fclose(fin);
   }

   ith=(int)(th);
   MPI_Bcast(&ith,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Nf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&order,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&iroot,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncfg,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&lambda,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   th=(theory_t)(ith);
}


static void default_pv(void)
{
   int i,pmax;
   double *p;

   pmax=(int)(PM_TYPES);
   p=malloc(pmax*(order+1)*sizeof(*p));

   for (i=0;i<(pmax*(order+1));i++)
      p[i]=0.0;

   for (i=0;i<pmax;i++)
   {
      opv[i]=order;
      pv[i]=p;
      p+=order+1;
   }

   pv[(int)(C0)][0]=1.0;
   pv[(int)(LAMBDA)][0]=1.0;
   pv[(int)(MASS)][0]=0.0;
   pv[(int)(CG)][0]=1.0;
   pv[(int)(CGP)][0]=1.0;
   pv[(int)(CSW)][0]=1.0;
   pv[(int)(CF)][0]=1.0;
   pv[(int)(CFP)][0]=1.0;
}


static void first_order(sl3_alg_dble *fld)
{
   sl3_alg_dble *asrc[2];
   sl3_alg_dble **wad;
   rflds_t *rf;

   wad=reserve_wad(1);
   rf=rflds();

   asrc[0]=(*rf).flds[(int)(GLUON)][0];
   asrc[1]=asrc[0];

   AGn(3,asrc,fld);
   mulr_calg(4*VOLUME,1.0/6.0,fld,wad[0]);
   aprop(wad[0],fld);

   release_wad();
}


static void second_order(sl3_alg_dble *fld)
{
   sl3_alg_dble *asrc[3],*cb,*c;
   sl3_alg_dble **wad,**wcd;
   rflds_t *rf;

   wad=reserve_wad(2);
   wcd=reserve_wcd(2);
   rf=rflds();

   asrc[0]=(*rf).flds[(int)(GLUON)][0];
   asrc[1]=asrc[0];
   asrc[2]=asrc[0];

   AGn(4,asrc,fld);
   mulr_calg(4*VOLUME,1.0/24.0,fld,wad[0]);

   if (th==LYM)
   {
      Amn(2,asrc,fld);
      mulr_add_calg(4*VOLUME,wad[0],fld,1.0/2.0);
   }

   AGn(3,asrc,fld);
   aprop(fld,wad[1]);
   asrc[1]=wad[1];
   AGn(3,asrc,fld);
   mulr_add_calg(4*VOLUME,wad[0],fld,5.0/72.0);

   c=(*rf).flds[(int)(GHOST)][1];
   cb=(*rf).flds[(int)(ANTIGHOST)][1];

   if (th==LYM)
   {
      AFPn(2,asrc,cb,c,fld);
      mulr_add_calg(4*VOLUME,wad[0],fld,-1.0/2.0);
   }

   if (iroot==0)
   {
      CFPn(1,asrc,c,wcd[0]);
      cprop(wcd[0],wcd[1]);
      AFPn(1,asrc,cb,wcd[1],fld);
      mulr_add_calg(4*VOLUME,wad[0],fld,-1.0/2.0);
   }
   else
   {
      CbFPn(1,asrc,cb,wcd[0]);
      cprop(wcd[0],wcd[1]);
      AFPn(1,asrc,wcd[1],c,fld);
      mulr_add_calg(4*VOLUME,wad[0],fld,-1.0/2.0);
   }

   asrc[1]=(*rf).flds[(int)(GLUON)][1];
   asrc[2]=asrc[1];
   AGn(4,asrc,fld);
   mulr_add_calg(4*VOLUME,wad[0],fld,1.0/8.0);

   AGn(3,asrc,fld);
   aprop(fld,wad[1]);
   asrc[0]=wad[1];
   AGn(3,asrc,fld);
   mulr_add_calg(4*VOLUME,wad[0],fld,1.0/12.0);

   aprop(wad[0],fld);

   release_wcd();
   release_wad();
}


int main(int argc,char *argv[])
{
   int i,k,icfg,icl,ncl,nw[4];
   double dev1,dev2;
   double wt1,wt2,wdt,wtav;
   sl3_alg_dble **wad0,**wad1;
   tupel_t *tp;
   rcls_t *rc;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   set_error_file("check3.log",0);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);

      printf("\n");
      printf("Direct check of ispt()\n");
      printf("----------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
      fflush(flog);
   }

   read_infile();
   set_th_parms(th,bc,Nf);
   default_pv();
   pv[(int)(LAMBDA)][0]=lambda;
   set_ispt_parms(order,iroot,opv,pv);
   print_th_parms();
   print_ispt_parms();

   start_ranlux(0,12345);
   geometry();

   for (i=0;i<4;i++)
      nw[i]=0;

   nw[1]=2*(order+1)+2;
   nw[2]=2;
   init_ispt(nw);
   wad0=reserve_wad(order+1);
   wad1=reserve_wad(order+1);
   rc=rtree_classes(th);
   ncl=rc[0].cls[0].n;

   for (icfg=0;icfg<ncfg;icfg++)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      generate_rflds(wad0[0]);

      for (icl=0;icl<ncl;icl++)
      {
         tp=rc[0].cls[0].tpa[icl];

         for (i=1;i<(*tp).n;i++)
            if ((*tp).i[i]!=0)
               break;

         if (i==(*tp).n)
         {
            k=(*tp).i[0];
            ispt(icl,wad0[k]);

            if (k==1)
               first_order(wad1[k]);

            if (k==2)
               second_order(wad1[k]);
         }
      }

      mulr_add_calg(4*VOLUME,wad1[1],wad0[1],-1.0);
      dev1=norm_square_calg(4*VOLUME,1,wad1[1])/
         norm_square_calg(4*VOLUME,1,wad0[1]);
      dev1=sqrt(dev1);

      mulr_add_calg(4*VOLUME,wad1[2],wad0[2],-1.0);
      dev2=norm_square_calg(4*VOLUME,1,wad1[2])/
         norm_square_calg(4*VOLUME,1,wad0[2]);
      dev2=sqrt(dev2);

      wt2=MPI_Wtime();
      wdt=wt2-wt1;
      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      wtav/=(double)(NPROC);

      if (my_rank==0)
      {
         printf("\n");
         printf("Configuration no %d, time %.1e sec\n\n",icfg,wtav);

         printf("Deviations:\n");
         printf("First order  = %.1e\n",dev1);
         printf("Second order = %.1e\n",dev2);

         printf("\n");
         fflush(flog);
      }
   }

   release_wad();
   release_wad();
   stop_ispt();

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
