
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs set_adbc() and check_adbc().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "sl3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "aflds.h"
#include "linalg.h"
#include "lattice.h"
#include "global.h"


static void new_flds(int n,sl3_alg_dble **ad)
{
   int i;

   for (i=0;i<n;i++)
   {
      if (cpr[0]==(NPROC0-1))
         gauss_dble((double*)(ad[i]),16*(4*VOLUME+7*(BNDRY/4)+3));
      else
         gauss_dble((double*)(ad[i]),16*(4*VOLUME+7*(BNDRY/4)));
   }
}


static int is_equal(sl3_alg_dble *ad1,sl3_alg_dble *ad2)
{
   int ie;

   ie=1;
   ie&=((*ad1).c1.re==(*ad2).c1.re);
   ie&=((*ad1).c1.im==(*ad2).c1.im);
   ie&=((*ad1).c2.re==(*ad2).c2.re);
   ie&=((*ad1).c2.im==(*ad2).c2.im);
   ie&=((*ad1).c3.re==(*ad2).c3.re);
   ie&=((*ad1).c3.im==(*ad2).c3.im);
   ie&=((*ad1).c4.re==(*ad2).c4.re);
   ie&=((*ad1).c4.im==(*ad2).c4.im);
   ie&=((*ad1).c5.re==(*ad2).c5.re);
   ie&=((*ad1).c5.im==(*ad2).c5.im);
   ie&=((*ad1).c6.re==(*ad2).c6.re);
   ie&=((*ad1).c6.im==(*ad2).c6.im);
   ie&=((*ad1).c7.re==(*ad2).c7.re);
   ie&=((*ad1).c7.im==(*ad2).c7.im);
   ie&=((*ad1).c8.re==(*ad2).c8.re);
   ie&=((*ad1).c8.im==(*ad2).c8.im);

   return ie;
}


static int is_zero(sl3_alg_dble *ad)
{
   int ie;

   ie=1;
   ie&=((*ad).c1.re==0.0);
   ie&=((*ad).c1.im==0.0);
   ie&=((*ad).c2.re==0.0);
   ie&=((*ad).c2.im==0.0);
   ie&=((*ad).c3.re==0.0);
   ie&=((*ad).c3.im==0.0);
   ie&=((*ad).c4.re==0.0);
   ie&=((*ad).c4.im==0.0);
   ie&=((*ad).c5.re==0.0);
   ie&=((*ad).c5.im==0.0);
   ie&=((*ad).c6.re==0.0);
   ie&=((*ad).c6.im==0.0);
   ie&=((*ad).c7.re==0.0);
   ie&=((*ad).c7.im==0.0);
   ie&=((*ad).c8.re==0.0);
   ie&=((*ad).c8.im==0.0);

   return ie;
}


static int cmp_active(sl3_alg_dble *ad1,sl3_alg_dble *ad2)
{
   int bc,ix,t,ifc,ie;

   bc=bc_type();
   ie=1;

   for (ix=(VOLUME/2);ix<(VOLUME);ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((ifc<2)||(t>0)||(bc==2))
            ie&=is_equal(ad1,ad2);

         ad1+=1;
         ad2+=1;
      }
   }

   return ie;
}


static int check_bval(sl3_alg_dble *ad)
{
   int bc,ix,t,ifc,ie;

   bc=bc_type();
   ie=1;

   for (ix=(VOLUME/2);ix<(VOLUME);ix++)
   {
      t=global_time(ix);

      if ((bc==1)&&(t==0))
      {
         for (ifc=0;ifc<8;ifc++)
         {
            if (ifc>1)
               ie&=is_zero(ad);
            ad+=1;
         }
      }
      else
         ad+=8;
   }

   if (cpr[0]==(NPROC0-1))
   {
      ad+=7*(BNDRY/4);
      ie&=is_zero(ad);
      ie&=is_zero(ad+1);
      ie&=is_zero(ad+2);
   }

   return ie;
}


int main(int argc,char *argv[])
{
   int my_rank,bc,ie;
   sl3_alg_dble **wad;
   th_parms_t thp;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);
      printf("\n");
      printf("Check of set_adbc() and check_adbc()\n");
      printf("------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check2.c]",
                    "Syntax: check2 [-bc <type>]");
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();

   thp=th_parms();
   error(thp.bc!=bc,1,"main [check2.c]",
         "Type of boundary condition is not properly set");

   alloc_wad(2);
   wad=reserve_wad(2);
   new_flds(2,wad);

   assign_calg2calg(4*VOLUME+7*(BNDRY/4),wad[0],wad[1]);
   set_adbc(wad[0]);

   ie=check_adbc(wad[0]);
   error(ie!=1,1,"main [check2.c]",
         "check_adbc() is incorrect");

   ie=check_adbc(wad[1]);
   error(ie!=0,2,"main [check2.c]",
         "check_adbc() is incorrect");

   ie=cmp_active(wad[0],wad[1]);
   error(ie!=1,1,"main [check2.c]",
         "Active link variables are modified by set_adbc()");

   ie=check_bval(wad[0]);
   error(ie!=1,1,"main [check2.c]",
         "Boundary values are not properly set by set_adbc()");

   release_wad();

   if (my_rank==0)
   {
      printf("No errors detected\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
