
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs set_cdbc() and check_cdbc().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "global.h"

static sl3_alg_dble ct0;


static void new_flds(int n,sl3_alg_dble **cd)
{
   int i;

   for (i=0;i<n;i++)
      gauss_dble((double*)(cd[i]),16*NSPIN);
}


static int is_equal(sl3_alg_dble *cd1,sl3_alg_dble *cd2)
{
   int ie;

   ie=1;
   ie&=((*cd1).c1.re==(*cd2).c1.re);
   ie&=((*cd1).c1.im==(*cd2).c1.im);
   ie&=((*cd1).c2.re==(*cd2).c2.re);
   ie&=((*cd1).c2.im==(*cd2).c2.im);
   ie&=((*cd1).c3.re==(*cd2).c3.re);
   ie&=((*cd1).c3.im==(*cd2).c3.im);
   ie&=((*cd1).c4.re==(*cd2).c4.re);
   ie&=((*cd1).c4.im==(*cd2).c4.im);
   ie&=((*cd1).c5.re==(*cd2).c5.re);
   ie&=((*cd1).c5.im==(*cd2).c5.im);
   ie&=((*cd1).c6.re==(*cd2).c6.re);
   ie&=((*cd1).c6.im==(*cd2).c6.im);
   ie&=((*cd1).c7.re==(*cd2).c7.re);
   ie&=((*cd1).c7.im==(*cd2).c7.im);
   ie&=((*cd1).c8.re==(*cd2).c8.re);
   ie&=((*cd1).c8.im==(*cd2).c8.im);

   return ie;
}


static void set_ct0(sl3_alg_dble *cd)
{
   int ip,iy;
   int cp[4];

   cp[0]=0;
   cp[1]=0;
   cp[2]=0;
   cp[3]=0;

   ip=ipr_global(cp);
   iy=ipt[0];
   ct0=cd[iy];

   MPI_Bcast(&ct0,16,MPI_DOUBLE,ip,MPI_COMM_WORLD);
}


static int check_cdbnd(sl3_alg_dble *cd)
{
   int ix,ie;

   ie=1;

   if (bc_type()==1)
   {
      set_ct0(cd);

      for (ix=0;ix<VOLUME;ix++)
      {
         if (global_time(ix)==0)
            ie&=is_equal(&ct0,cd+ix);
      }
   }

   return ie;
}


static int cmp_cd(sl3_alg_dble *cd1,sl3_alg_dble *cd2)
{
   int ix,ie;

   ie=1;

   if (bc_type()==1)
   {
      for (ix=0;ix<VOLUME;ix++)
      {
         if (global_time(ix)!=0)
            ie&=is_equal(cd1+ix,cd2+ix);
      }
   }
   else
   {
      for (ix=0;ix<VOLUME;ix++)
         ie&=is_equal(cd1+ix,cd2+ix);
   }

   for (ix=VOLUME;ix<NSPIN;ix++)
      ie&=is_equal(cd1+ix,cd2+ix);

   return ie;
}


int main(int argc,char *argv[])
{
   int my_rank,bc,ie;
   sl3_alg_dble **wcd;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);
      printf("\n");
      printf("Check of set_cdbc() and check_cdbc()\n");
      printf("------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check3.c]",
                    "Syntax: check3 [-bc <type>]");
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();

   alloc_wcd(3);
   wcd=reserve_wcd(2);
   new_flds(2,wcd);

   assign_calg2calg(NSPIN,wcd[0],wcd[1]);
   set_cdbc(wcd[0]);

   ie=check_cdbnd(wcd[0]);
   error(ie!=1,1,"main [check3.c]",
         "Boundary values are not correctly set");

   ie=check_cdbc(wcd[0]);
   error(ie!=1,1,"main [check3.c]",
         "check_cdbc() is not correct ");

   if (bc==1)
   {
      ie=check_cdbc(wcd[1]);
      error(ie!=0,2,"main [check3.c]",
            "check_cdbc() is not correct ");
   }

   ie=cmp_cd(wcd[0],wcd[1]);
   error(ie!=1,1,"main [check3.c]",
         "set_cdbc() changed field where it should not");

   release_wcd();

   if (my_rank==0)
   {
      printf("No errors detected\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
