
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Checks of the programs in the module sl3alg.c.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "random.h"
#include "su3fcts.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "global.h"

#define NALG 100033

static int *rlxs_state=NULL,*rlxd_state=NULL;
static double var[64],var_all[64];


static int is_zero(sl3_alg_dble *ad)
{
   int ie;

   ie=1;
   ie&=((*ad).c1.re==0.0);
   ie&=((*ad).c1.im==0.0);
   ie&=((*ad).c2.re==0.0);
   ie&=((*ad).c2.im==0.0);
   ie&=((*ad).c3.re==0.0);
   ie&=((*ad).c3.im==0.0);
   ie&=((*ad).c4.re==0.0);
   ie&=((*ad).c4.im==0.0);
   ie&=((*ad).c5.re==0.0);
   ie&=((*ad).c5.im==0.0);
   ie&=((*ad).c6.re==0.0);
   ie&=((*ad).c6.im==0.0);
   ie&=((*ad).c7.re==0.0);
   ie&=((*ad).c7.im==0.0);
   ie&=((*ad).c8.re==0.0);
   ie&=((*ad).c8.im==0.0);

   return ie;
}


static void save_ranlux(void)
{
   int nlxs,nlxd;

   if (rlxs_state==NULL)
   {
      nlxs=rlxs_size();
      nlxd=rlxd_size();

      rlxs_state=malloc((nlxs+nlxd)*sizeof(int));
      rlxd_state=rlxs_state+nlxs;

      error(rlxs_state==NULL,1,"save_ranlux [ms1.c]",
            "Unable to allocate state arrays");
   }

   rlxs_get(rlxs_state);
   rlxd_get(rlxd_state);
}


static void restore_ranlux(void)
{
   rlxs_reset(rlxs_state);
   rlxd_reset(rlxd_state);
}


int main(int argc,char *argv[])
{
   int my_rank,n,i,j,ie;
   double dev,dmax,dmax_all;
   double nsq1,nsq2;
   double sm,r[8];
   double rn,cij,eij;
   complex_dble z1,z2;
   su3_dble *M,*m,w;
   sl3_alg_dble *X,*Y,*x;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);

      printf("\n");
      printf("Checks of the programs in the module sl3alg.c\n");
      printf("---------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      printf("Number of field components: %d\n\n",NALG);
   }

   start_ranlux(0,98765);
   geometry();

   X=amalloc(2*NALG*sizeof(*X),ALIGN);
   M=amalloc(NALG*sizeof(*M),ALIGN);
   error((X==NULL)||(M==NULL),1,
         "main [check2.c]","Unable to allocate field arrays");
   Y=X+NALG;

   set_calg2zero(NALG,X);
   ie=1;

   for (n=0;n<NALG;n++)
   {
      ie&=is_zero(X+n);
      X[n].c3.re=1.0;
   }

   error(ie!=1,1,"main [check2.c]","set_calg2zero() is not correct");
   dmax=fabs(norm_square_calg(NALG,1,X)-2.0*(double)(NALG*NPROC));
   MPI_Reduce(&dmax,&dmax_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {
      printf("Check of norm_square_calg():\n");
      printf("Element count = %.1e (should be 0.0)\n",dmax_all);
   }

   sm=0.0;
   dmax=0.0;

   for (n=0;n<NALG;n++)
   {
      x=X+n;
      m=M+n;

      gauss_dble((double*)(x),16);

      (*m).c11.re=(*x).c1.re;
      (*m).c11.im=(*x).c1.im;
      (*m).c22.re=(*x).c2.re;
      (*m).c22.im=(*x).c2.im;
      (*m).c33.re=-(*x).c1.re-(*x).c2.re;
      (*m).c33.im=-(*x).c1.im-(*x).c2.im;

      (*m).c12.re= (*x).c3.re;
      (*m).c12.im= (*x).c3.im;
      (*m).c21.re=-(*x).c4.re;
      (*m).c21.im= (*x).c4.im;

      (*m).c13.re= (*x).c5.re;
      (*m).c13.im= (*x).c5.im;
      (*m).c31.re=-(*x).c6.re;
      (*m).c31.im= (*x).c6.im;

      (*m).c23.re= (*x).c7.re;
      (*m).c23.im= (*x).c7.im;
      (*m).c32.re=-(*x).c8.re;
      (*m).c32.im= (*x).c8.im;

      su3dagxsu3(m,m,&w);
      nsq1=norm_square_calg(1,0,x);
      nsq2=2.0*(w.c11.re+w.c22.re+w.c33.re);

      dev=fabs(1.0-nsq2/nsq1);
      if (dev>dmax)
         dmax=dev;

      sm+=nsq2;
   }

   MPI_Reduce(&dmax,&dmax_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {
      printf("|1.0+2*tr{X^2}/||X||^2| = %.1e (single elements)\n",dmax_all);
      printf("(should be less than %.1e or so)\n",DBL_EPSILON*sqrt(8.0));
   }

   dmax=fabs(1.0-sm/norm_square_calg(NALG,0,X));
   MPI_Reduce(&dmax,&dmax_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {
      printf("|1.0+2*tr{X^2}/||X||^2| = %.1e (whole vector)\n",dmax_all);
      printf("(should be less than %.1e or so)\n\n",
             DBL_EPSILON*sqrt(8.0*(double)(NALG)));
   }

   gauss_dble((double*)(X),16*NALG);
   assign_calg2calg(NALG,X,Y);

   for (n=0;n<NALG;n++)
   {
      _sl3_alg_sub_assign(Y[n],X[n]);
   }

   nsq2=norm_square_calg(NALG,1,Y);

   if (my_rank==0)
   {
      printf("Check of assign_calg2calg(): %.1e\n",nsq2);
      printf("(should be 0.0)\n\n");
   }

   z1.re=cos(0.123)-1.0;
   z1.im=sin(0.123);
   gauss_dble((double*)(X),16*NALG);
   assign_calg2calg(NALG,X,Y);
   mulc_add_calg(NALG,Y,X,z1);
   nsq1=norm_square_calg(NALG,1,X);
   nsq2=norm_square_calg(NALG,1,Y);

   if (my_rank==0)
      printf("Check of mulc_add_calg(): %.1e\n\n",fabs(nsq2/nsq1-1.0));

   z1.re=0.123;
   z1.im=-4.567;
   gauss_dble((double*)(X),16*NALG);
   gauss_dble((double*)(Y),16*NALG);
   nsq1=norm_square_calg(NALG,1,X);
   nsq2=norm_square_calg(NALG,1,Y);
   z2=scalar_prod_calg(NALG,1,X,Y);
   mulc_add_calg(NALG,Y,X,z1);
   nsq1=nsq2+2.0*(z1.re*z2.re+z1.im*z2.im)+(z1.re*z1.re+z1.im*z1.im)*nsq1;
   nsq2=norm_square_calg(NALG,1,Y);

   if (my_rank==0)
      printf("Check of scalar_prod_calg(): %.1e\n\n",fabs(nsq2/nsq1-1.0));

   random_calg(NALG,X);

   for (i=0;i<8;i++)
   {
      for (j=0;j<8;j++)
         var[8*i+j]=0.0;
   }

   for (n=0;n<NALG;n++)
   {
      r[0]=(X[n].c1.im-X[n].c2.im)/3.0;
      r[1]=(X[n].c2.im+2.0*X[n].c1.im)/3.0;
      r[2]=X[n].c3.re;
      r[3]=X[n].c3.im;
      r[4]=X[n].c5.re;
      r[5]=X[n].c5.im;
      r[6]=X[n].c7.re;
      r[7]=X[n].c7.im;

      for (i=0;i<8;i++)
      {
         for (j=i;j<8;j++)
            var[8*i+j]+=r[i]*r[j];
      }
   }

   MPI_Reduce(var,var_all,64,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);

   if (my_rank==0)
   {
      printf("Check of random_calg():\n\n");
      dmax=0.0;
      rn=1.0/((double)(NALG)*(double)(NPROC));

      for (i=0;i<8;i++)
      {
         for (j=i;j<8;j++)
         {
            if ((i==j)&&(i>1))
            {
               cij=1.0/4.0;
               eij=sqrt(2.0*rn)/4.0;
            }
            else if (i==j)
            {
               cij=1.0/9.0;
               eij=sqrt(2.0*rn)/9.0;
            }
            else if ((i==0)&&(j==1))
            {
               cij=1.0/18.0;
               eij=sqrt(5.0*rn)/18.0;
            }
            else if ((i<2)&&(j>1))
            {
               cij=0.0;
               eij=sqrt(rn)/6.0;
            }
            else
            {
               cij=0.0;
               eij=sqrt(rn)/4.0;
            }

            var_all[8*i+j]*=rn;

            if (cij!=0.0)
            {
               printf("<b[%d]*b[%d]> = % .4e, deviation = %.1e+-%.1e\n",
                      i,j,var_all[8*i+j],fabs(var_all[8*i+j]-cij),eij);
            }
            else
            {
               dev=fabs(var_all[8*i+j])/eij;

               if (dev>dmax)
                  dmax=dev;
            }
         }
      }

      eij=sqrt(rn)/4.0;
      printf("\n");
      printf("For all other i,j, ");
      printf("max|<b[i]*b[j]>| = %.1e (should be %.1e or so)\n\n",
             dmax*eij,2.0*eij);
   }

   save_ranlux();

   for (n=0;n<NALG;n++)
   {
      random_calg(1,X+n);
      random_calg(1,Y+n);
   }

   z1.re=0.0;
   z1.im=-1.0;
   mulc_add_calg(NALG,X,Y,z1);

   restore_ranlux();

   random_calg2(NALG,Y);
   nsq1=norm_square_calg(NALG,1,Y);
   mulr_add_calg(NALG,Y,X,-1.0/sqrt(2.0));
   nsq2=norm_square_calg(NALG,1,Y);

   if (my_rank==0)
   {
      printf("Check of random_calg2():\n\n");
      printf("Deviation from random_calg() = %.1e\n",sqrt(nsq2/nsq1));
      nsq1/=(double)(8.0*NALG*NPROC);
      printf("<|X|^2> = % .4e, deviation = %.1e+-%.1e\n\n",
             nsq1,fabs(nsq1-1.0),sqrt(1.0/(double)(8*NALG*NPROC)));
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
