
/*******************************************************************************
*
* File table2.c
*
* Copyright (C) 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Table of the normalization factor of the gradient-flow coupling (see
* modules/nompi/gbar/kfact.c and doc/gfnorm.pdf).
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "utils.h"
#include "gbar.h"

static int ns,*Ts,*Ls,bcs,Ns;
static double c1s,cs;
static char type[NAME_SIZE];


static void read_parms(void)
{
   int i,ie;
   FILE *fin=NULL;

   fin=fopen("table2.in","r");
   error(fin==NULL,1,"read_parms [table2.c]",
         "Unable to open input parameter file");

   find_section(fin,"Lattice");
   ns=count_tokens(fin,"T");
   error((ns<1)||(ns!=count_tokens(fin,"L")),1,"read_parms [table2.c]",
         "Improper lists of T and L values");
   Ts=malloc(2*ns*sizeof(*Ts));
   Ls=Ts+ns;
   error(Ts==NULL,1,"read_parms [table2.c]",
         "Unable to allocate size arrays");
   read_iprms(fin,"T",ns,Ts);
   read_iprms(fin,"L",ns,Ls);
   read_line(fin,"bc","%d",&bcs);

   find_section(fin,"Theory");
   read_line(fin,"N","%d",&Ns);
   read_line(fin,"c1","%lf\n",&c1s);

   find_section(fin,"Coupling");
   read_line(fin,"type","%s\n",type);
   read_line(fin,"c","%lf\n",&cs);

   ie=0;

   for (i=0;i<ns;i++)
      ie|=((Ts[i]<4)||((Ts[i]%2)!=0)||(Ls[i]<4));

   error((ie!=0)||(bcs<1)||(bcs>2)||(Ns<=1)||(c1s>=0.125)||(cs<=0.0),1,
         "read_parms [table2.c]","Improper choice of input parameters");
   error((strcmp(type,"clover")!=0)&&(strcmp(type,"plaquette")!=0),1,
         "read_parms [table2.c]","Parameter type must be clover or plaquette");

   fclose(fin);
}


static void print_table(FILE *fdat)
{
   int n;
   double c0,t,k,ks,kt;
   char ty0[2],ty1[3],ty2[3];

   ty0[0]=type[0];
   ty1[0]=type[0];
   ty2[0]=type[0];
   ty1[1]='s';
   ty2[1]='t';
   ty0[1]='\0';
   ty1[2]='\0';
   ty2[2]='\0';

   fprintf(fdat,"#\n");
   fprintf(fdat,"# Normalization factors k_%s,k_%s,k_%s\n",ty0,ty1,ty2);
   fprintf(fdat,"#\n");

   if (bcs==1)
      fprintf(fdat,"# SF boundary conditions\n");
   else
      fprintf(fdat,"# Open-SF boundary conditions\n");

   fprintf(fdat,"# N = %d\n",Ns);
   c0=1.0-8.0*c1s;
   n=fdigits(c0);
   fprintf(fdat,"# c0 = %.*f, ",IMAX(n,1),c0);
   n=fdigits(c1s);
   fprintf(fdat,"c1 = %.*f\n",IMAX(n,1),c1s);
   n=fdigits(cs);
   fprintf(fdat,"# c = %.*f\n",IMAX(n,1),cs);
   fprintf(fdat,"#\n");

   fprintf(fdat,"#  T   L      k_%s       k_%s       k_%s",ty0,ty1,ty2);
   fprintf(fdat,"       1/k_%s        1/k_%s       1/k_%s\n",ty0,ty1,ty2);

   for (n=0;n<ns;n++)
   {
      t=cs*(double)(Ls[n]);
      t=(t*t)/8.0;
      k=kfact(ty0,Ns,Ts[n],Ls[n],bcs,c1s,Ts[n]/2,t);
      ks=kfact(ty1,Ns,Ts[n],Ls[n],bcs,c1s,Ts[n]/2,t);
      kt=kfact(ty2,Ns,Ts[n],Ls[n],bcs,c1s,Ts[n]/2,t);

      fprintf(fdat," %3d %3d %10.5f %10.5f %10.5f ",Ts[n],Ls[n],k,ks,kt);
      fprintf(fdat,"%12.9f %12.9f %12.9f\n",1.0/k,1.0/ks,1.0/kt);
   }
}


int main(int argc,char *argv[])
{
   FILE *fdat;

   read_parms();
   fdat=fopen("plots/table2.dat","w");
   error(fdat==NULL,1,"main [table2.c]","Unable to open output file");
   print_table(fdat);
   fclose(fdat);

   printf("\n");
   printf("Normalization factor k_%s of the gradient-flow coupling\n\n",
          type);
   printf("Table printed to file plots/table2.dat\n\n");
   exit(0);
}
