
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generation of Wick contractions.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "utils.h"
#include "permutations.h"

static int **ia,*ip,(*ps)[2];


static void chk_npm(int m,int nb,int *n,long npm)
{
   int k,l;

   for (k=0;k<m;k++)
   {
      if (k<nb)
      {
         for (l=1;l<n[k];l+=2)
            npm/=(long)(l);
      }
      else
      {
         for (l=1;l<=(n[k]/2);l++)
            npm/=(long)(l);
      }
   }

   error(npm!=1L,1,"chk_npm [check2.c]","Unexpected number of pairings");
}


static int chk_pairs(int m,int nb,int *n,int **a,int (*p)[2])
{
   int k,l,j;

   for (k=0;k<m;k++)
   {
      for (l=0;l<n[k];l++)
         ia[k][l]=0;
   }

   for (k=0;k<m;k++)
   {
      for (l=0;l<n[k];l+=2)
      {
         if (k<nb)
         {
            for (j=0;j<n[k];j++)
            {
               if ((p[0][0]==a[k][j])&&(ia[k][j]==0))
               {
                  ia[k][j]=1;
                  break;
               }
            }

            if (j==n[k])
               return 1;

            for (j=0;j<n[k];j++)
            {
               if ((p[0][1]==a[k][j])&&(ia[k][j]==0))
               {
                  ia[k][j]=1;
                  break;
               }
            }

            if (j==n[k])
               return 1;

            p+=1;
         }
         else
         {
            for (j=0;j<(n[k]/2);j++)
            {
               if ((p[0][0]==a[k][j])&&(ia[k][j]==0))
               {
                  ia[k][j]=1;
                  break;
               }
            }

            if (j==n[k])
               return 1;

            for (j=(n[k]/2);j<n[k];j++)
            {
               if ((p[0][1]==a[k][j])&&(ia[k][j]==0))
               {
                  ia[k][j]=1;
                  break;
               }
            }

            if (j==n[k])
               return 1;

            p+=1;
         }
      }
   }

   for (k=0;k<m;k++)
   {
      for (l=0;l<n[k];l++)
      {
         if (ia[k][l]!=1)
            return 1;
      }
   }

   return 0;
}


static int cmp_pair(int n,int (*p1)[2],int (*p2)[2])
{
   int i,j;

   for (i=0;i<n;i++)
      ip[i]=0;

   for (i=0;i<n;i++)
   {
      for (j=0;j<n;j++)
      {
         if (ip[j]==0)
         {
            if (((p1[i][0]==p2[j][0])&&(p1[i][1]==p2[j][1]))||
                ((p1[i][0]==p2[j][1])&&(p1[i][1]==p2[j][0])))
            {
               ip[j]=1;
               break;
            }
         }
      }

      if (j==n)
         return 1;
   }

   for (i=0;i<n;i++)
   {
      if (ip[i]!=1)
         return 1;
   }

   return 0;
}


static int chk_all(int m,int nb,int *n,int **a,long np,int (*p)[2])
{
   int k;
   int (*p1)[2],(*p2)[2];
   long nps;

   stop_wick();

   nps=0L;
   (void)(init_wick(m,nb,n,a));

   while (nps<(np-1L))
   {
      nps=next_wick(ps);
      p1=p;
      p2=ps;

      for (k=0;k<m;k++)
      {
         if (cmp_pair(n[k]/2,p1,p2))
            break;

         p1+=n[k]/2;
         p2+=n[k]/2;
      }

      if (k==m)
         return 1;
   }

   nps=next_wick(ps);

   return 0;
}


int main(void)
{
   int nb,nf,m,nt,i,j,k;
   int *n,*aa,**a,(*p)[2];
   long np,npm;

   printf("\n");
   printf("Generation of Wick contractions\n");
   printf("-------------------------------\n\n");

   while(1)
   {
      printf("Specify number of bosonic and fermionic subsequences: ");
      (void)scanf("%d %d",&nb,&nf);

      m=nb+nf;
      n=malloc(m*sizeof(*n));
      error(n==NULL,1,"main [check2.c]","Unable to allocate lengths array");

      printf("Specify lengths of subsequences: ");

      for (i=0;i<m;i++)
         (void)scanf("%d",n+i);

      printf("\n");
      printf("nb = %d, nf = %d, n = %d",nb,nf,n[0]);
      nt=n[0];

      for (i=1;i<m;i++)
      {
         printf(",%d",n[i]);
         nt+=n[i];
      }

      printf("\n\n");

      a=malloc(2*m*sizeof(*a));
      aa=malloc((2*nt+nt/2)*sizeof(*aa));
      error((a==NULL)||(aa==NULL),1,"main [check2.c]",
            "Unable to allocate index array");

      ia=a+m;
      a[0]=aa;
      ia[0]=aa+nt;
      ip=aa+2*nt;

      for (i=0;i<m;i++)
      {
         if (i>0)
         {
            a[i]=a[i-1]+n[i-1];
            ia[i]=ia[i-1]+n[i-1];
         }

         printf("Specify subsequence no %d (%d elements): ",i,n[i]);

         for (j=0;j<n[i];j++)
            (void)scanf("%d",a[i]+j);
      }

      printf("Wick contractions:\n");

      np=0L;
      npm=init_wick(m,nb,n,a);
      chk_npm(m,nb,n,npm);

      printf("Total number of contractions = %ld\n",npm);

      p=malloc(nt*sizeof(*p));
      error(p==NULL,1,"main [check2.c]","Unable to allocate pair array");
      ps=p+nt;

      while (np<npm)
      {
         np=next_wick(p);
         printf("np = %3ld, ",np);
         k=0;

         for (i=0;i<m;i++)
         {
            if (i==0)
               printf("[");
            else
               printf(",[");

            for (j=0;j<(n[i]/2);j++)
            {
               if (j==0)
                  printf("(%d,%d)",p[k][0],p[k][1]);
               else
                  printf(",(%d,%d)",p[k][0],p[k][1]);
               k+=1;
            }

            printf("]");
         }

         printf("\n");

         error(chk_pairs(m,nb,n,a,p),1,"main [check2.c]",
               "Element mismatch");

         error(chk_all(m,nb,n,a,np,p),1,"main [check2.c]",
               "Dublicated contractions");
      }

      printf("\n");
      printf("Wrapping around:\n");
      np=0L;

      while (np<(npm/2))
      {
         np=next_wick(p);
         printf("np = %3ld, ",np);
         k=0;

         for (i=0;i<m;i++)
         {
            if (i==0)
               printf("[");
            else
               printf(",[");

            for (j=0;j<(n[i]/2);j++)
            {
               if (j==0)
                  printf("(%d,%d)",p[k][0],p[k][1]);
               else
                  printf(",(%d,%d)",p[k][0],p[k][1]);
               k+=1;
            }

            printf("]");
         }

         printf("\n");
      }

      printf("...\n\n");
      stop_wick();
   }

   exit(0);
}
