
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Drawing of labeled tree diagrams.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "utils.h"
#include "plot.h"
#include "contract.h"

static vlist_t vls;


static theory_t theory(char *name)
{
   if (strcmp("PHI4",name)==0)
      return PHI4;
   else if (strcmp("YM",name)==0)
      return YM;
   else if (strcmp("QCD",name)==0)
      return QCD;
   else if (strcmp("LYM",name)==0)
      return LYM;
   else if (strcmp("LQCD",name)==0)
      return LQCD;
   else
      error(1,1,"theory [check3.c]","Unknown theory");

   return THEORIES;
}


static int nogfix(tlist_t *tl)
{
   int n,i,nd;
   ltree_t **lta;

   n=(*tl).n;
   lta=(*tl).lta;
   nd=0;

   for (i=0;i<n;i++)
   {
      if (nVX(lta[i][0].dm,VGF,&vls)==0)
         nd+=1;
   }

   return nd;
}


static int noghost(tlist_t *tl)
{
   int n,i,nd;
   ltree_t **lta;

   n=(*tl).n;
   lta=(*tl).lta;
   nd=0;

   for (i=0;i<n;i++)
   {
      if ((nVX(lta[i][0].dm,VGF,&vls)==0)&&
          (nVX(lta[i][0].dm,VFP,&vls)==0))
         nd+=1;
   }

   return nd;
}


int main(void)
{
   int order,nlp;
   int io,il,is,ic;
   int ip,nc,n,i;
   double pos1[2],pos2[2];
   char name[5],plot_file[NAME_SIZE];
   theory_t th;
   cnum_t **cna;
   ltree_t **lta;
   tlist_t *tl;
   FILE *plt;

   printf("\n");
   printf("Drawing of labeled tree diagrams\n");
   printf("--------------------------------\n\n");

   while (1)
   {
      printf("Specify theory and maximal order: ");
      (void)scanf("%4s %d",name,&order);
      printf("Specify root line position (0|1): ");
      (void)scanf("%d",&is);
      printf("Specify whether coefficients should be drawn (0|1): ");
      (void)scanf("%d",&ic);

      th=theory(name);
      tmap(th,order,is);
      vls=vertices(th,order);

      printf("Number of labeled tree diagrams:\n");
      printf("  order    nloop     ntree\n");

      for (io=1;io<=order;io++)
      {
         nlp=nlp_max(th,io);

         for (il=0;il<=nlp;il++)
         {
            tl=select_ltrees(th,io,il);
            printf("    %d        %d     %5d\n",
                   io,il,(*tl).n);
         }
      }

      printf("\n");
      printf("Print these diagrams? (1|0) ");
      (void)scanf("%d",&ip);

      if (ip==1)
      {
         printf("Number of columns = ");
         (void)scanf("%d",&nc);
         if (nc<1)
            nc=1;

         sprintf(plot_file,"plots/%s.%d.s4.tex",name,order);
         plt=fopen(plot_file,"w");
         error(plt==NULL,1,"main [check3.c]","Unable to open plot file");
         start_plot(plt,nc);
         vls=vertices(th,order);

         for (io=1;io<=order;io++)
         {
            nlp=nlp_max(th,io);

            for (il=0;il<=nlp;il++)
            {
               tl=select_ltrees(th,io,il);

               if (tl!=NULL)
               {
                  n=(*tl).n;
                  lta=(*tl).lta;
                  cna=(*tl).cna;

                  for (i=0;i<n;i++)
                  {
                     start_bbox();
                     if (ic==1)
                        draw_ltree1(cna[i],lta[i],&vls);
                     else
                        draw_ltree0(lta[i],&vls);
                     end_bbox(pos1,pos2);

                     start_diagram(pos1,pos2);
                     if (ic==1)
                        draw_ltree1(cna[i],lta[i],&vls);
                     else
                        draw_ltree0(lta[i],&vls);
                     end_diagram();
                  }
               }
            }
         }

         end_plot();
         fclose(plt);
         printf("Printed plot commands to file %s\n\n",plot_file);
      }

      if (th!=PHI4)
      {
         printf("\n");
         printf("Number of labeled tree diagrams w/o gauge-fixing "
                "vertices:\n");
         printf("  order    nloop     ntree\n");

         for (io=1;io<=order;io++)
         {
            nlp=nlp_max(th,io);

            for (il=0;il<=nlp;il++)
            {
               tl=select_ltrees(th,io,il);
               printf("    %d        %d     %5d\n",
                      io,il,nogfix(tl));
            }
         }

         printf("\n");
         printf("Print these diagrams? (1|0) ");
         (void)scanf("%d",&ip);

         if (ip==1)
         {
            printf("Number of columns = ");
            (void)scanf("%d",&nc);
            if (nc<1)
               nc=1;

            sprintf(plot_file,"plots/%s.%d.s4.ngfix.tex",name,order);
            plt=fopen(plot_file,"w");
            error(plt==NULL,1,"main [check3.c]","Unable to open plot file");
            start_plot(plt,nc);
            vls=vertices(th,order);

            for (io=1;io<=order;io++)
            {
               nlp=nlp_max(th,io);

               for (il=0;il<=nlp;il++)
               {
                  tl=select_ltrees(th,io,il);

                  if (tl!=NULL)
                  {
                     n=(*tl).n;
                     lta=(*tl).lta;
                     cna=(*tl).cna;

                     for (i=0;i<n;i++)
                     {
                        if (nVX(lta[i][0].dm,VGF,&vls)==0)
                        {
                           start_bbox();
                           if (ic==1)
                              draw_ltree1(cna[i],lta[i],&vls);
                           else
                              draw_ltree0(lta[i],&vls);
                           end_bbox(pos1,pos2);

                           start_diagram(pos1,pos2);
                           if (ic==1)
                              draw_ltree1(cna[i],lta[i],&vls);
                           else
                              draw_ltree0(lta[i],&vls);
                           end_diagram();
                        }
                     }
                  }
               }
            }

            end_plot();
            fclose(plt);
            printf("Printed plot commands to file %s\n\n",plot_file);
         }

         printf("\n");
         printf("Number of labeled tree diagrams w/o ghosts & gauge-fixing "
                "vertices:\n");
         printf("  order    nloop     ntree\n");

         for (io=1;io<=order;io++)
         {
            nlp=nlp_max(th,io);

            for (il=0;il<=nlp;il++)
            {
               tl=select_ltrees(th,io,il);
               printf("    %d        %d     %5d\n",
                      io,il,noghost(tl));
            }
         }

         printf("\n");
         printf("Print these diagrams? (1|0) ");
         (void)scanf("%d",&ip);

         if (ip==1)
         {
            printf("Number of columns = ");
            (void)scanf("%d",&nc);
            if (nc<1)
               nc=1;

            sprintf(plot_file,"plots/%s.%d.s4.nghost.tex",name,order);
            plt=fopen(plot_file,"w");
            error(plt==NULL,1,"main [check3.c]","Unable to open plot file");
            start_plot(plt,nc);
            vls=vertices(th,order);

            for (io=1;io<=order;io++)
            {
               nlp=nlp_max(th,io);

               for (il=0;il<=nlp;il++)
               {
                  tl=select_ltrees(th,io,il);

                  if (tl!=NULL)
                  {
                     n=(*tl).n;
                     lta=(*tl).lta;
                     cna=(*tl).cna;

                     for (i=0;i<n;i++)
                     {
                        if ((nVX(lta[i][0].dm,VGF,&vls)==0)&&
                            (nVX(lta[i][0].dm,VFP,&vls)==0))
                        {
                           start_bbox();
                           if (ic==1)
                              draw_ltree1(cna[i],lta[i],&vls);
                           else
                              draw_ltree0(lta[i],&vls);
                           end_bbox(pos1,pos2);

                           start_diagram(pos1,pos2);
                           if (ic==1)
                              draw_ltree1(cna[i],lta[i],&vls);
                           else
                              draw_ltree0(lta[i],&vls);
                           end_diagram();
                        }
                     }
                  }
               }
            }

            end_plot();
            fclose(plt);
            printf("Printed plot commands to file %s\n\n",plot_file);
         }
      }
   }

   exit(0);
}
