
/*******************************************************************************
*
* File check4.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the program get_rtree().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "utils.h"
#include "trees.h"
#include "rvt.h"
#include "ispt.h"

static int naw=0,nvw=0,nlw=0;
static int iv,il,na,nv,nl;
static int nlegs[(int)(LTYPES)],*pv;
static vlist_t vls;
static ltree_t *ltw;


static theory_t theory(char *name)
{
   if (strcmp("PHI4",name)==0)
      return PHI4;
   else if (strcmp("YM",name)==0)
      return YM;
   else if (strcmp("QCD",name)==0)
      return QCD;
   else if (strcmp("LYM",name)==0)
      return LYM;
   else if (strcmp("LQCD",name)==0)
      return LQCD;
   else
      error(1,1,"theory [check4.c]","Unknown theory");

   return THEORIES;
}


static void check_vx(rtree_t *rt)
{
   int nsrc,nbr;
   int i,it,ir,ie;
   vertex_t *vx;
   ltype_t lt;
   src_t *src;
   rtree_t *br;

   nsrc=(*rt).nsrc;
   nbr=(*rt).nbr;
   lt=(*rt).lt;
   src=(*rt).src;
   br=(*rt).br;

   nv+=1;
   nl+=nbr;

   vx=vls.v[(*rt).v[0]]+(*rt).v[1];

   for (i=0;i<(int)(LTYPES);i++)
      nlegs[i]=(*vx).nl[i];

   nlegs[(int)(lt)]-=1;

   for (i=0;i<nsrc;i++)
   {
      it=(int)(src[i].lt);
      ir=(int)(rvt[it]);
      nlegs[ir]-=1;

      if (src[i].isrc!=0)
         na+=1;
   }

   for (i=0;i<nbr;i++)
   {
      it=(int)(br[i].lt);
      ir=(int)(rvt[it]);
      nlegs[ir]-=1;
   }

   ie=0;

   for (i=0;i<(int)(LTYPES);i++)
      ie|=(nlegs[i]!=0);

   error(ie!=0,1,"check_vx [check4.c]","Incorrect vertex legs");

   for (i=0;i<nbr;i++)
      check_vx(br+i);
}


static void alloc_ltw(void)
{
   diagram_t *dm;

   if ((nv>nvw)||(nl>nlw))
   {
      if ((nvw>0)||(nlw>0))
      {
         free(pv);
         free_dgm((*ltw).dm);
      }

      if (nv>nvw)
         nvw=nv;
      if (nl>nlw)
         nlw=nl;

      pv=malloc(nvw*sizeof(*pv));
      dm=alloc_dgm(nvw,nlw);
      error(pv==NULL,1,"alloc_ltw [check4.c]",
            "Unable to allocate auxiliary array");
   }
   else
      dm=(*ltw).dm;

   if (na>naw)
   {
      if (naw>0)
         free_ltd(ltw);

      naw=na;
      ltw=alloc_ltd(naw);
   }

   (*ltw).nl=na;
   (*ltw).dm=dm;
   (*dm).nv=nv;
   (*dm).nl=nl;
}


static int set_vx(rtree_t *rt)
{
   int nsrc,nbr,i;
   int iv0,il0,(*v)[2];
   ltype_t lt;
   line_t *l;
   diagram_t *dm;
   src_t *src;
   rtree_t *br;

   nsrc=(*rt).nsrc;
   src=(*rt).src;

   for (i=0;i<nsrc;i++)
   {
      if (src[i].isrc>0)
      {
         lt=src[i].lt;

         if (lt==ANTIGHOST)
            lt=GHOST;
         if (lt==ANTIQUARK)
            lt=QUARK;

         nlegs[(int)(lt)]+=1;
      }
   }

   nbr=(*rt).nbr;
   br=(*rt).br;

   dm=(*ltw).dm;
   v=(*dm).v;
   l=(*dm).l;

   iv0=iv;
   iv+=1;

   v[iv0][0]=(*rt).v[0];
   v[iv0][1]=(*rt).v[1];

   for (i=0;i<nbr;i++)
   {
      il0=il;
      il+=1;

      l[il0].type=br[i].lt;
      l[il0].iv[1]=iv0;
      l[il0].iv[0]=set_vx(br+i);
   }

   return iv0;
}


static int set_la(rtree_t *rt)
{
   int nsrc,nbr;
   int it,ie,i;
   ltype_t lt;
   line_t *la;
   src_t *src;
   rtree_t *br;

   nsrc=(*rt).nsrc;
   nbr=(*rt).nbr;
   src=(*rt).src;
   br=(*rt).br;
   ie=0;

   for (i=0;i<nsrc;i++)
   {
      if (src[i].isrc>0)
      {
         lt=src[i].lt;

         if ((lt==ANTIGHOST)||(lt==ANTIQUARK))
            it=(int)(rvt[(int)(lt)]);
         else
            it=(int)(lt);

         la=(*ltw).l+nlegs[it]+src[i].isrc-1;
         (*la).type=(ltype_t)(it);

         if ((lt==ANTIGHOST)||(lt==ANTIQUARK))
            (*la).iv[0]=iv;
         else if ((lt==GHOST)||(lt==QUARK))
            (*la).iv[1]=iv;
         else
         {
            if ((*la).iv[0]==nv)
               (*la).iv[0]=iv;
            else if ((*la).iv[1]==nv)
               (*la).iv[1]=iv;
            else
               ie=1;
         }
      }
   }

   iv+=1;

   for (i=0;i<nbr;i++)
      ie|=set_la(br+i);

   return ie;
}


static int chk_la(void)
{
   int i,ie,iv;
   line_t *la;

   ie=0;
   la=(*ltw).l;

   for (i=0;i<na;i++)
   {
      ie|=(la[i].type==LTYPES);
      iv=la[i].iv[0];
      ie|=((iv<0)||(iv>=nv));
      iv=la[i].iv[1];
      ie|=((iv<0)||(iv>=nv));
   }

   return ie;
}


static void set_ltw(rtree_t *rt)
{
   int i,ie,sm,osm;

   for (i=0;i<(int)(LTYPES);i++)
      nlegs[i]=0;

   for (i=0;i<na;i++)
   {
      (*ltw).l[i].iv[0]=nv;
      (*ltw).l[i].iv[1]=nv;
   }

   iv=0;
   il=0;
   (void)(set_vx(rt));
   ie=0;
   sm=0;

   for (i=0;i<(int)(LTYPES);i++)
   {
      ie|=((nlegs[i]%2)!=0);
      nlegs[i]/=2;
      osm=sm;
      sm+=nlegs[i];
      nlegs[i]=osm;
   }

   error(ie!=0,1,"set_ltw [check4.c]","Labeled external lines are not paired");

   ie=is_tree((*ltw).dm);
   error(ie!=1,1,"set_ltw [check4.c]","Failed to reconstruct diagram");

   iv=0;
   ie=set_la(rt);
   ie|=chk_la();

   error(ie!=0,1,"set_ltw [check4.c]","Inconsistent labeled external lines");
}


static void rpv(line_t *rl,line_t *l)
{
   int i,iv0,iv1;
   line_t rlb;

   iv0=(*rl).iv[0];
   iv1=(*rl).iv[1];

   rlb.type=(*rl).type;
   rlb.iv[1]=iv0;
   pv[iv]=iv0;
   iv+=1;

   for (i=0;i<nl;i++)
   {
      if ((l[i].iv[0]==iv0)&&(l[i].iv[1]!=iv1))
      {
         rlb.iv[0]=l[i].iv[1];
         rpv(&rlb,l);
      }
      else if ((l[i].iv[1]==iv0)&&(l[i].iv[0]!=iv1))
      {
         rlb.iv[0]=l[i].iv[0];
         rpv(&rlb,l);
      }
   }
}


static void set_pv(theory_t th,ltree_t *ltd)
{
   line_t rl;
   diagram_t *dm;

   if (th==PHI4)
      rl.type=SCALAR;
   else
      rl.type=GLUON;

   rl.iv[0]=(*ltd).rv;
   rl.iv[1]=nv;

   iv=0;
   dm=(*ltd).dm;

   rpv(&rl,(*dm).l);
}


static void chk_vertices(diagram_t *dm1,diagram_t *dm2)
{
   int i,j,ie;
   int (*v1)[2],(*v2)[2];

   v1=(*dm1).v;
   v2=(*dm2).v;
   ie=0;

   for (i=0;i<nv;i++)
   {
      j=pv[i];

      ie|=((v1[i][0]!=v2[j][0])||(v1[i][1]!=v2[j][1]));
   }

   error(ie!=0,1,"chk_vertices [check4.c]","Incorrect vertex list");
}


static int cmp_line(line_t *l1,line_t *l2)
{
   if ((pv[(*l1).iv[0]]==(*l2).iv[0])&&(pv[(*l1).iv[1]]==(*l2).iv[1])&&
       ((*l1).type==(*l2).type))
      return 1;

   if ((pv[(*l1).iv[0]]==(*l2).iv[1])&&(pv[(*l1).iv[1]]==(*l2).iv[0])&&
       ((*l1).type==rvt[(int)((*l2).type)]))
      return 1;

   return 0;
}


static void chk_lines(diagram_t *dm1,diagram_t *dm2)
{
   int i,j,ie;
   line_t *l1,*l2;

   l1=(*dm1).l;
   l2=(*dm2).l;
   ie=0;

   for (i=0;i<nl;i++)
   {
      for (j=0;j<nl;j++)
      {
         if (cmp_line(l1+j,l2+i))
         {
            l1[j].type=LTYPES;
            break;
         }
      }

      ie|=(j==nl);
   }

   error(ie!=0,1,"chk_lines [check4.c]","Incorrect line list");
}


static void chk_loop_lines(line_t *la1,line_t *la2)
{
   int i,j,k;
   int ni,nj,ie;

   ie=0;

   for (i=0;i<na;i++)
   {
      if (la1[i].type==ANTIGHOST)
      {
         la1[i].type=GHOST;
         k=la1[i].iv[1];
         la1[i].iv[1]=la1[i].iv[0];
         la1[i].iv[0]=k;
      }

      if (la1[i].type==ANTIQUARK)
      {
         la1[i].type=QUARK;
         k=la1[i].iv[1];
         la1[i].iv[1]=la1[i].iv[0];
         la1[i].iv[0]=k;
      }

      ie|=((la2[i].type==ANTIGHOST)||(la2[i].type==ANTIQUARK));
   }

   error(ie!=0,1,"check_loop_lines [check4.c]",
         "There are antifermion loop lines");

   for (i=0;i<na;i++)
   {
      ni=0;

      for (k=0;k<i;k++)
         ni+=(la2[i].type==la2[k].type);

      for (j=0;j<na;j++)
      {
         if (cmp_line(la1+j,la2+i))
         {
            nj=0;

            for (k=0;k<j;k++)
               nj+=(la1[j].type==la1[k].type);

            if (ni==nj)
               break;
         }
      }

      ie|=(j==na);
   }

   error(ie!=0,1,"chk_loop_lines [check4.c]","Incorrect loop-line list");
}


int main(void)
{
   int order,is,nlp;
   int na0,na1,io,ilp;
   int n,i;
   char name[5];
   theory_t th;
   diagram_t *dm;
   ltree_t **lta;
   rtree_t *rt;
   tlist_t *tl;

   printf("\n");
   printf("Check of the program get_rtree()\n");
   printf("--------------------------------\n\n");

   ltw=alloc_ltd(0);

   while (1)
   {
      printf("Specify theory, maximal order and root selection flag: ");
      (void)(scanf("%4s %d %d",name,&order,&is));

      th=theory(name);
      na0=trees(th,order);
      na1=ltrees(th,order);
      set_root(th,is);

      printf("\n");
      printf("Total number of basic tree diagrams =   %d\n",na0);
      printf("Total number of labeled tree diagrams = %d\n\n",na1);

      printf("Numbers of tree diagrams of a given order:\n");
      printf("  order    nloop     ntree\n");

      for (io=1;io<=order;io++)
      {
         nlp=nlp_max(th,io);

         for (ilp=0;ilp<=nlp;ilp++)
         {
            tl=select_ltrees(th,io,ilp);
            printf("    %d        %d     %5d\n",
                   io,ilp,(*tl).n);
         }
      }

      vls=vertices(th,order);

      for (io=1;io<=order;io++)
      {
         nlp=nlp_max(th,io);

         for (ilp=0;ilp<=nlp;ilp++)
         {
            tl=select_ltrees(th,io,ilp);
            n=(*tl).n;
            lta=(*tl).lta;

            for (i=0;i<n;i++)
            {
               dm=lta[i][0].dm;
               rt=get_rtree(th,order,lta[i]);
               na=0;
               nv=0;
               nl=0;
               check_vx(rt);

               error(((na%2)!=0)||(lta[i][0].nl!=(na/2))||
                     (nv!=(*dm).nv)||(nl!=(*dm).nl),1,"main [check4.c]",
                     "Incorrect vertex or line numbers");

               na/=2;
               alloc_ltw();
               set_ltw(rt);
               set_pv(th,lta[i]);
               chk_vertices((*ltw).dm,dm);
               chk_lines((*ltw).dm,dm);
               chk_loop_lines((*ltw).l,lta[i][0].l);
            }
         }
      }

      printf("\n");
   }

   exit(0);
}
