
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the program sym_prod().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "utils.h"
#include "random.h"
#include "permutations.h"
#include "su3fcts.h"
#include "vxfcts.h"

static int *pa;
static su3_dble *X=NULL,*PX,*Y,*PY,*W;


static double max_dev(su3_dble *u,su3_dble *v)
{
   int i;
   double r[18],s[18];
   double nrm,d,dmax;

   r[ 0]=(*u).c11.re;
   r[ 1]=(*u).c11.im;
   r[ 2]=(*u).c12.re;
   r[ 3]=(*u).c12.im;
   r[ 4]=(*u).c13.re;
   r[ 5]=(*u).c13.im;

   r[ 6]=(*u).c21.re;
   r[ 7]=(*u).c21.im;
   r[ 8]=(*u).c22.re;
   r[ 9]=(*u).c22.im;
   r[10]=(*u).c23.re;
   r[11]=(*u).c23.im;

   r[12]=(*u).c31.re;
   r[13]=(*u).c31.im;
   r[14]=(*u).c32.re;
   r[15]=(*u).c32.im;
   r[16]=(*u).c33.re;
   r[17]=(*u).c33.im;

   s[ 0]=(*v).c11.re;
   s[ 1]=(*v).c11.im;
   s[ 2]=(*v).c12.re;
   s[ 3]=(*v).c12.im;
   s[ 4]=(*v).c13.re;
   s[ 5]=(*v).c13.im;

   s[ 6]=(*v).c21.re;
   s[ 7]=(*v).c21.im;
   s[ 8]=(*v).c22.re;
   s[ 9]=(*v).c22.im;
   s[10]=(*v).c23.re;
   s[11]=(*v).c23.im;

   s[12]=(*v).c31.re;
   s[13]=(*v).c31.im;
   s[14]=(*v).c32.re;
   s[15]=(*v).c32.im;
   s[16]=(*v).c33.re;
   s[17]=(*v).c33.im;

   nrm=0.0;
   dmax=0.0;

   for (i=0;i<18;i++)
   {
      nrm+=r[i]*r[i];
      d=(r[i]-s[i])*(r[i]-s[i]);
      i+=1;
      nrm+=r[i]*r[i];
      d+=(r[i]-s[i])*(r[i]-s[i]);

      if (d>dmax)
         dmax=d;
   }

   return sqrt(dmax/nrm);
}


static double chk_sym_prod(int type,int n,unsigned int idx)
{
   int i,p;
   unsigned int id;
   double r;
   long np,ip;

   if (type==0)
   {
      p=0;
      id=1u;

      for (i=0;i<n;i++)
      {
         if (id&idx)
         {
            Y[p]=X[i];
            p+=1;
         }

         id<<=1u;
      }

      cm3x3_zero(1,PY+idx);
      np=init_perm(1,&p);
      ip=0L;

      while (ip<np)
      {
         ip=next_perm(pa);
         cm3x3_assign(1,Y+pa[0],W);

         for (i=1;i<p;i++)
            su3xsu3(Y+pa[i],W,W);

         cm3x3_add(W,PY+idx);
      }

      stop_perm();

      r=1.0/(double)(np);
   }
   else
   {
      p=bcount(n,idx);

      if (type==1)
         r=1.0/(double)(factorial(p));
      else
         r=1.0;

      if ((p&0x1)&&((type==1)||(type==3)))
         r=-r;
   }

   cm3x3_mulr(&r,PY+idx,W);

   return max_dev(PX+idx,W);
}


int main(void)
{
   int type,n;
   unsigned int idx,idm,imx;
   double d,dmax;

   printf("\n");
   printf("Check of the program sym_prod()\n");
   printf("-------------------------------\n\n");

   while (1)
   {
      printf("Specify number of factors: ");
      (void)(scanf("%d",&n));
      printf("\n");

      error((n<0)||(n>=(8*sizeof(unsigned int))),1,"main [check2.c]",
            "Number of factors is out of range");

      if (X!=NULL)
      {
         free(pa);
         afree(X);
      }

      imx=1u<<n;
      pa=malloc(n*sizeof(*pa));
      X=amalloc((2*(n+imx)+1)*sizeof(*X),4);
      error((pa==NULL)||(X==NULL),1,"main [check2.c]",
            "Unable to allocate matrix array");
      PX=X+n;
      Y=PX+imx;
      PY=Y+n;
      W=PY+imx;
      gauss_dble((double*)(X),36*(n+imx));

      for (type=0;type<4;type++)
      {
         sym_prod(type,n,X,PX);
         dmax=0.0;
         idm=0u;

         for (idx=0u;idx<imx;idx++)
         {
            if (idx>0u)
               d=chk_sym_prod(type,n,idx);
            else
            {
               cm3x3_unity(1,W);
               d=max_dev(W,PX);
            }

            if (d>dmax)
            {
               idm=idx;
               dmax=d;
            }
         }

         printf("Maximal deviation (type=%d) = %.1e at idx = %#X\n",
                type,dmax,idm);
      }

      printf("\n");
   }

   exit(0);
}
