
/*******************************************************************************
*
* File check1.c
*
* Copyright (C) 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the program plaq_frc().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "sl3.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "su3fcts.h"
#include "aflds.h"
#include "uflds.h"
#include "wflow.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define NMAX 4

static complex_dble dSplaq[NMAX+1] ALIGNED16;
static complex_dble s0[NMAX+1] ALIGNED16;
static complex_dble s1[NMAX+1] ALIGNED16;
static complex_dble sw[NMAX+1] ALIGNED16;
static su3_dble Xw ALIGNED16;


static void new_fld(sl3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),16*(4*VOLUME+7*(BNDRY/4)+3));
   else
      gauss_dble((double*)(ad),16*(4*VOLUME+7*(BNDRY/4)));

   set_adbc(ad);
}


static void copy_ud(int n,su3_dble *ud,su3_dble *sv)
{
   int ix;

   for (ix=0;ix<(4*VOLUME*(n+1));ix++)
      sv[ix]=ud[ix];
}


static void rot_ud(int n,double eps,sl3_alg_dble *ad,su3_dble *ud)
{
   int ix,i;

   for (ix=0;ix<(4*VOLUME);ix++)
   {
      sl3_to_su3(ad,&Xw);
      cm3x3_mulr(&eps,&Xw,&Xw);

      for (i=n;i>0;i--)
         cm3x3_mul_add(&Xw,ud+i-1,ud+i);

      cm3x3_add(&Xw,ud);
      ud+=(n+1);
      ad+=1;
   }
}


static void dSdt(int n,sl3_alg_dble *ad,sl3_alg_dble *fd)
{
   int ix,t,ifc,i;
   int ism;

   ism=init_hsum(2*(n+1));
   reset_hsum(ism);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         for (i=0;i<=n;i++)
         {
            dSplaq[i]=sl3_biform(ad,fd);

            if ((t==0)&&(ifc>1))
            {
               dSplaq[i].re*=0.5;
               dSplaq[i].im*=0.5;
            }

            fd+=1;
         }

         add_to_hsum(ism,(double*)(dSplaq));
         ad+=1;
      }
   }

   global_hsum(ism,(double*)(dSplaq));
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,i;
   double eps,d,dmax;
   complex_dble z;
   sl3_alg_dble **wad;
   su3_dble **wud,**wsv;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check1.log","w",stdout);

      printf("\n");
      printf("Check of the program plaq_frc()\n");
      printf("-------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check1.c]",
                    "Syntax: check1 [-bc <type>]");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();
   alloc_wad(NMAX+2);
   alloc_wud(2*(NMAX+1));
   wad=reserve_wad(NMAX+2);
   wud=reserve_wud(NMAX+1);
   wsv=reserve_wud(NMAX+1);

   for (n=0;n<=NMAX;n++)
   {
      for (i=0;i<=n;i++)
         new_fld(wad[i]);

      a2uflds(n,wad,wud[0]);
      copy_ud(n,wud[0],wsv[0]);
      plaq_frc(n,wud[0],wad[0]);
      new_fld(wad[n+1]);
      dSdt(n,wad[n+1],wad[0]);

      eps=2.0e-4;
      rot_ud(n,eps,wad[n+1],wud[0]);
      Splaq(n,wud[0],s0);

      for (i=0;i<=n;i++)
      {
         s0[i].re*=(2.0/3.0);
         s0[i].im*=(2.0/3.0);
      }

      copy_ud(n,wsv[0],wud[0]);
      rot_ud(n,-eps,wad[n+1],wud[0]);
      Splaq(n,wud[0],s1);

      for (i=0;i<=n;i++)
      {
         s1[i].re*=(2.0/3.0);
         s1[i].im*=(2.0/3.0);
      }

      copy_ud(n,wsv[0],wud[0]);
      rot_ud(n,2.0*eps,wad[n+1],wud[0]);
      Splaq(n,wud[0],sw);

      for (i=0;i<=n;i++)
      {
         s0[i].re-=(sw[i].re/12.0);
         s0[i].im-=(sw[i].im/12.0);
      }

      copy_ud(n,wsv[0],wud[0]);
      rot_ud(n,-2.0*eps,wad[n+1],wud[0]);
      Splaq(n,wud[0],sw);

      for (i=0;i<=n;i++)
      {
         s1[i].re-=(sw[i].re/12.0);
         s1[i].im-=(sw[i].im/12.0);
      }

      for (i=0;i<=n;i++)
      {
         sw[i].re=(s0[i].re-s1[i].re)/eps;
         sw[i].im=(s0[i].im-s1[i].im)/eps;
      }

      if (my_rank==0)
      {
         dmax=0.0;

         for (i=0;i<=n;i++)
         {
            z.re=dSplaq[i].re+sw[i].re;
            z.im=dSplaq[i].im+sw[i].im;
            d=z.re*z.re+z.im*z.im;

            z.re=dSplaq[i].re;
            z.im=dSplaq[i].im;
            d/=(z.re*z.re+z.im*z.im);

            if (d>dmax)
               dmax=d;
         }

         printf("n = %d, maximal relative deviation = %.1e\n",n,sqrt(dmax));
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
