
/*******************************************************************************
*
* File check4.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of t^2*<E>.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "mpi.h"
#include "pmdb.h"
#include "utils.h"
#include "random.h"
#include "lattice.h"
#include "uflds.h"
#include "tupel.h"
#include "linalg.h"
#include "wflow.h"
#include "ispt.h"
#include "global.h"

static int my_rank,bc,Nf,order,iroot,ncfg;
static int ns,idmp,rule;
static int opv[(int)(PM_TYPES)];
static double lambda,eps,*pv[(int)(PM_TYPES)];
static double *plaq_sq;
static complex_dble *plaq_sp,*plaq_tm,*plaq_sum;
static theory_t th;
static char theory[NAME_SIZE];
static FILE *fin=NULL,*flog=NULL;


static void set_th(void)
{
   if (strcmp("PHI4",theory)==0)
      th=PHI4;
   else if (strcmp("YM",theory)==0)
      th=YM;
   else if (strcmp("QCD",theory)==0)
      th=QCD;
   else if (strcmp("LYM",theory)==0)
      th=LYM;
   else if (strcmp("LQCD",theory)==0)
      th=LQCD;
   else
      error(1,1,"set_th [check4.c]","Unknown theory");
}


static void read_theory_parms(void)
{
   find_section("Theory");
   read_line("theory","%s",theory);
   read_line("bc","%d",&bc);
   read_line("Nf","%d",&Nf);
   set_th();

   find_section("ISPT");
   read_line("order","%d",&order);
   read_line("iroot","%d",&iroot);
   read_line("lambda","%lf",&lambda);

   find_section("Run");
   read_line("ncfg","%d",&ncfg);
}


static void read_wflow_parms(void)
{
   find_section("Wilson flow");
   read_line("ns","%d",&ns);
   read_line("eps","%lf",&eps);
   read_line("idmp","%d",&idmp);
   read_line("rule","%d",&rule);
}


static void read_infile(void)
{
   int ith;

   if (my_rank==0)
   {
      fin=freopen("check4.in","r",stdin);
      error_root(fin==NULL,1,"read_infile [check4.c]",
                 "Unable to open input file");
      read_theory_parms();
      read_wflow_parms();
      fclose(fin);
   }

   ith=(int)(th);
   MPI_Bcast(&ith,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Nf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&order,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&iroot,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&lambda,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncfg,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ns,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&eps,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&idmp,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rule,1,MPI_INT,0,MPI_COMM_WORLD);
   th=(theory_t)(ith);
}


static void default_pv(void)
{
   int i,pmax;
   double *p;

   pmax=(int)(PM_TYPES);
   p=malloc(pmax*(order+1)*sizeof(*p));

   for (i=0;i<(pmax*(order+1));i++)
      p[i]=0.0;

   for (i=0;i<pmax;i++)
   {
      opv[i]=order;
      pv[i]=p;
      p+=order+1;
   }

   pv[(int)(C0)][0]=1.0;
   pv[(int)(LAMBDA)][0]=1.0;
   pv[(int)(MASS)][0]=0.0;
   pv[(int)(CG)][0]=1.0;
   pv[(int)(CGP)][0]=1.0;
   pv[(int)(CSW)][0]=1.0;
   pv[(int)(CF)][0]=1.0;
   pv[(int)(CFP)][0]=1.0;
}


static void alloc_arrays(void)
{
   plaq_sq=malloc((order+1)*sizeof(*plaq_sq));
   plaq_sp=malloc(3*(order+1)*sizeof(*plaq_sp));
   error((plaq_sq==NULL)||(plaq_sp==NULL),1,"alloc_arrays [check4.c]",
         "Unable to allocate data arrays");

   plaq_tm=plaq_sp+order+1;
   plaq_sum=plaq_tm+order+1;
}


int main(int argc,char *argv[])
{
   int i,k,icfg,icl,ncl,nw[4];
   double r,tsq,wt1,wt2,wdt,wtav;
   sl3_alg_dble **wad;
   su3_dble **wud;
   tupel_t *tp;
   rcls_t *rc;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   set_error_file("check4.log",0);

   if (my_rank==0)
   {
      flog=freopen("check4.log","w",stdout);

      printf("\n");
      printf("Computation of t^2*<E> (plaquette definition)\n");
      printf("---------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
      fflush(flog);
   }

   read_infile();
   set_th_parms(th,bc,Nf);
   default_pv();
   pv[(int)(LAMBDA)][0]=lambda;
   set_ispt_parms(order,iroot,opv,pv);
   print_th_parms();
   print_ispt_parms();

   if (my_rank==0)
   {
      printf("ns = %d\n",ns);
      printf("eps = %.3e\n",eps);
      printf("idmp = %d\n",idmp);

      if (rule==1)
         printf("Using the Euler integrator\n\n");
      else if (rule==2)
         printf("Using the 2nd order RK integrator\n\n");
      else
         printf("Using the 3rd order RK integrator\n\n");

      fflush(flog);
   }

   start_ranlux(0,12345);
   geometry();

   for (i=0;i<4;i++)
      nw[i]=0;

   nw[1]=3*(order+1);
   nw[2]=order+1;
   init_ispt(nw);
   alloc_wud(order+1);
   alloc_arrays();

   rc=rtree_classes(th);
   ncl=rc[0].cls[0].n;

   for (k=0;k<=order;k++)
   {
      plaq_sum[k].re=0.0;
      plaq_sum[k].im=0.0;
      plaq_sq[k]=0.0;
   }

   r=(double)(ns)*eps;
   tsq=-18.0*r*r;

   for (icfg=0;icfg<ncfg;icfg++)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      wad=reserve_wad(order+1);
      generate_rflds(wad[0]);

      for (icl=0;icl<ncl;icl++)
      {
         tp=rc[0].cls[0].tpa[icl];

         for (i=1;i<(*tp).n;i++)
            if ((*tp).i[i]!=0)
               break;

         if (i==(*tp).n)
         {
            k=(*tp).i[0];
            ispt(icl,wad[k]);
         }
      }

      wud=reserve_wud(order+1);
      a2uflds(order,wad,wud[0]);
      release_wad();

      if (rule==1)
         fwd_euler(ns,eps,idmp,order,wud[0]);
      else if (rule==2)
         fwd_rk2(ns,eps,idmp,order,wud[0]);
      else
         fwd_rk3(ns,eps,idmp,order,wud[0]);

      Eplaq(order,wud[0],plaq_sp,plaq_tm);
      release_wud();

      for (k=0;k<=order;k++)
      {
         plaq_sp[k].re+=plaq_tm[k].re;
         plaq_sp[k].im+=plaq_tm[k].im;
         plaq_sp[k].re*=tsq;
         plaq_sp[k].im*=tsq;

         plaq_sum[k].re+=plaq_sp[k].re;
         plaq_sum[k].im+=plaq_sp[k].im;

         plaq_sq[k]+=plaq_sp[k].re*plaq_sp[k].re;
      }

      wt2=MPI_Wtime();
      wdt=wt2-wt1;
      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      wtav/=(double)(NPROC);

      if (my_rank==0)
      {
         printf("\n");
         printf("Configuration no %d, time %.1e sec\n\n",icfg+1,wtav);

         for (k=0;k<=order;k++)
            printf("t^2*E[%d] = (% .5e,% .5e)\n",
                   k,plaq_sp[k].re,plaq_sp[k].im);

         printf("\n");
         fflush(flog);
      }
   }

   if (my_rank==0)
   {
      printf("Summary\n");
      printf("-------\n\n");

      printf("Number of configurations = %d\n",ncfg);
      r=1.0/(double)(ncfg);

      for (k=0;k<=order;k++)
      {
         plaq_sum[k].re*=r;
         plaq_sum[k].im*=r;
         plaq_sq[k]*=r;
      }

      for (k=0;k<=order;k++)
      {
         plaq_sq[k]-=plaq_sum[k].re*plaq_sum[k].re;
         plaq_sq[k]*=r;

         printf("t^2*E[%d] = (% .5e,% .5e), error = %.1e\n",
                k,plaq_sum[k].re,plaq_sum[k].im,sqrt(plaq_sq[k]));
      }

      printf("\n");
      fflush(flog);
   }

   stop_ispt();

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
