
/*******************************************************************************
*
* File diagrams.h
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
*******************************************************************************/

#ifndef DIAGRAMS_H
#define DIAGRAMS_H

typedef enum
{
   PHI4,YM,LYM,QCD,LQCD,
   THEORIES
} theory_t;

typedef enum
{
   SCALAR,GLUON,GHOST,ANTIGHOST,QUARK,ANTIQUARK,
   LTYPES
} ltype_t;

typedef enum
{
   SOURCE,EXTERNAL,
   VS,VG,VGM,VFP,VQG,
   VSM,VGF,VQM,
   VTYPES
} vtype_t;

typedef struct
{
   vtype_t type;
   int order;
   int nl[(int)(LTYPES)];
} vertex_t;

typedef struct
{
   int order;
   int *nv;
   vertex_t **v;
} vlist_t;

typedef struct
{
   ltype_t type;
   int iv[2];
} line_t;

typedef struct
{
   int nv,nl;
   int (*v)[2];
   line_t *l;
} diagram_t;

/* DIAGRAM_C */
extern diagram_t *alloc_dgm(int nv,int nl);
extern void free_dgm(diagram_t *dm);
extern void copy_dgm(diagram_t *dm1,diagram_t *dm2);
extern void remove_line(int i,diagram_t *dm1,diagram_t *dm2);
extern void connected_part(int i,diagram_t *dm1,diagram_t *dm2);

/* DMPROP_C */
extern int check_dgm(theory_t th,int order,diagram_t *dm);
extern int is_tree(diagram_t *dm);
extern int is_branch(diagram_t *dm);
extern int ncparts(diagram_t *dm);
extern int nquark_loops(diagram_t *dm);
extern int nghost_loops(diagram_t *dm);
extern int nlext(diagram_t *dm,vlist_t *vl,int *n1pt);
extern int has_tadpole(diagram_t *dm,ltype_t lt,vlist_t *vl);
extern int nVX(diagram_t *dm,vtype_t VX,vlist_t *vl);

/* ORDER_C */
extern int *alloc_rank(int nv);
extern void reorder_dgm(diagram_t *dm,int *rk);
extern int rank_dgm(diagram_t *dm1,int *rk1,diagram_t *dm2,int *rk2);
extern int cmp_dgm(diagram_t *dm1,int *rk1,diagram_t *dm2,int *rk2);

/* SFACTOR_C */
extern long sfactor(theory_t th,diagram_t *dm,long *sf);

/* VERTICES_C */
extern vlist_t vertices(theory_t th,int order);

#endif
